import importlib.util
import logging
import os
from typing import List

import numpy as np
from osgeo import gdal

import wapordl.toolbox.ogr_gdal as ogr_gdal

optional_packages = ["matplotlib"]
use_plt = all(
    [not isinstance(importlib.util.find_spec(x), type(None)) for x in optional_packages]
)
if use_plt:
    import matplotlib.patches
    import matplotlib.pyplot as plt

gdal.UseExceptions()


def geot_area(
    shape_fh: str,
    geot: List[float],
    zero_is_nan: bool = True,
    make_plots: bool | str = False,
) -> float:
    """Given a shape and a geotransform, this functions calculates the area of the pixels
    that would cover the shape.

    Parameters
    ----------
    shape_fh : str
        Path to a vector-file than can be opened by `gdal.OpenEx`.
    geot : List[float]
        A geotransform describing the locations of pixels of a raster.
    zero_is_nan : bool, optional
        Set the output to `np.nan` if the resulting area is zero, by default True.
    make_plots : bool | str, optional
        Create a plot visualising the shape and the overlapping raster. If a path to an existing
        folder is given, the plot is saved as a png file, by default False.

    Returns
    -------
    float
        Area of the pixels that would overlap with the shape.
    """
    global use_plt

    # Get the bounding-box of the shape
    bounds = ogr_gdal.get_bounds(shape_fh)
    coords = np.array(bounds).reshape((2, 2))

    # List which pixels intersect with the bb.
    nx_ny = np.array(
        [
            np.floor((coords[:, 0] - geot[0]) / geot[1]),
            np.ceil((coords[:, 1] - geot[3]) / abs(geot[5])),
        ]
    ).T

    # Snap the bounding-box to the geot.
    bounds = np.array(
        [
            geot[0] + nx_ny[0, 0] * geot[1],
            geot[3] + (nx_ny[0, 1] - 1) * abs(geot[5]),
            geot[0] + (nx_ny[-1, 0] + 1) * geot[1],
            geot[3] + nx_ny[-1, 1] * abs(geot[5]),
        ]
    ).T

    # Allow in-memory files (required for GDAL>=3.10)
    gdal_config_options = {"GDAL_MEM_ENABLE_OPEN": "YES"}

    # Set Rasterize options.
    rast_options = gdal.RasterizeOptions(
        burnValues=1,
        outputBounds=bounds,
        xRes=geot[1],
        yRes=geot[1],
        format="MEM",
        # allTouched=True,
    )

    # Make an array with 0=outside shape and 1=inside shape.
    try:
        for k, v in gdal_config_options.items():
            gdal.SetConfigOption(k, v)
        x = gdal.Rasterize("", shape_fh, options=rast_options)
    except Exception as e:
        raise e
    finally:
        for k, v in gdal_config_options.items():
            gdal.SetConfigOption(k, None)

    # Determine the area of the rasterizs shape.
    band = x.GetRasterBand(1)
    array = band.ReadAsArray()
    array[array == 0.0] = np.nan
    area = np.nansum(array) * geot[1] ** 2

    # Set to nan if requested.
    if zero_is_nan and area == 0.0:
        area = np.nan

    if make_plots and not use_plt:
        logging.info(
            "Unable to create plots without `matplotlib`, consider installing it or setting `make_plots=False`."
        )
    elif make_plots and area not in [0.0, np.nan] and use_plt:
        fig = plt.figure()
        ax = fig.gca()

        ax.imshow(
            array,
            cmap="tab10",
            extent=[bounds[0], bounds[2], bounds[1], bounds[3]],
            zorder=0,
        )

        geom, _ = ogr_gdal.get_geom(shape_fh)[:2]
        wkt = geom.UnaryUnion().ExportToWkt()
        coords = ogr_gdal.wkt_polygon_to_coords(wkt)
        shape_patch = matplotlib.patches.Polygon(
            coords,
            color="tab:red",
            zorder=10,
            alpha=0.4,
        )
        ax.add_patch(shape_patch)

        xticks_minor = np.arange(bounds[0], bounds[2], geot[1])
        xticks = np.arange(
            bounds[0] + 0.5 * geot[1], bounds[2] - 0.5 * geot[1], 3 * geot[1]
        )
        ax.set_xticks(xticks_minor, minor=True)
        ax.set_xticks(xticks)
        yticks_minor = np.arange(bounds[1], bounds[3], geot[1])
        yticks = np.arange(
            bounds[1] + 0.5 * geot[1], bounds[3] - 0.5 * geot[1], 3 * geot[1]
        )
        ax.set_yticks(yticks_minor, minor=True)
        ax.set_yticks(yticks)
        ax.grid(which="minor", color="w", linestyle=":", linewidth=1)

        ax.set_facecolor("lightgray")
        ax.set_xlabel("longitude [DD]")
        ax.set_ylabel("latitude [DD]")
        ax.set_title(f"FILE = {os.path.split(shape_fh)[-1]} \n\n PIXELSIZE = {geot[1]}")
        ax.tick_params(which="minor", bottom=False, left=False)

        if os.path.isdir(make_plots):
            plot_fh = os.path.join(
                make_plots,
                f"{area}_{os.path.split(shape_fh)[-1].replace('.geojson', '')}.png",
            )
            fig.savefig(plot_fh)
        plt.close(fig)
    else:
        ...

    return area


def determine_overview(
    info_fh: str,
    shape_fh: str | List[float],
    max_error: float = 0.5,
    make_plots: bool | str = False,
) -> int:
    """Given a raster file, iterates over the geotransforms of the raster and any
    overviews present and checks how well the pixels are able to represent a given shape.

    Parameters
    ----------
    info_fh : str
        Path to a raster file.
    shape_fh : str | List[float]
        Path to a vector file or a list describing a bounding-box as
        [xmin, ymin, xmax, ymax].
    max_error : float, optional
        Threshold for the iteration, by default 0.5.
    make_plots : bool | str, optional
        Create graphs showing the iteration process executed to determine the
        optimal overview. When an existing folder is defined, the graphs
        are saved into that folder as png files, by default False.

    Returns
    -------
    int
        The selected overview.
    """
    global use_plt

    # Load raster information.
    info = ogr_gdal.get_info(info_fh)
    epsg = int(info["coordinateSystem"]["wkt"].split('ID["EPSG",')[-1][:-2])

    if isinstance(shape_fh, list):
        shape_fh = ogr_gdal.to_vsimem(bb=shape_fh)
    if epsg != 4326:
        shape_fh = ogr_gdal.reproject_vector(shape_fh, epsg=epsg, in_memory=True)

    shape_area = ogr_gdal.get_area(shape_fh)

    # Determine the scales to convert the original geot.
    res = np.array(info["size"])
    overview_scales = {
        i: np.round(res / np.array(x["size"]))
        for i, x in enumerate(info["bands"][0]["overviews"])
    }
    overview_scales[-1] = np.array([1.0, 1.0])

    # Variables to store outputs.
    errors = list()
    overviews = list()

    # Loop over the outputs, starting with the coarsest.
    for overview, scales in sorted(overview_scales.items(), reverse=True):
        # Make overview geotransform.
        geot = info["geoTransform"] * np.array([1, scales[0], 1, 1, 1, scales[1]])
        # Determine the area of the pixels overlapping with the shape.
        outline_area = geot_area(shape_fh, geot, make_plots=make_plots)
        # Calculate the relative difference in area.
        error = abs(1 - (shape_area / outline_area)) * 100
        # Store outputs.
        errors.append(error)
        overviews.append(overview)
        # Stop when the error is below the threshold.
        if errors[-1] < max_error:
            break
        # Stop when the error doesn't decrease anymore.
        if len(errors) > 2:
            if errors[-3] - errors[-1] == 0.0:
                logging.info("Search converged.")
                break

    # Select the overview with the smallest error.
    if all(np.isnan(errors)):
        overview = -1
    else:
        overview = overviews[np.nanargmin(errors)]

    ogr_gdal.unlink_vsimems(shape_fh)

    # Create a plot if necessary.
    if make_plots and not use_plt:
        logging.info(
            "Unable to create plots without `matplotlib`, consider installing it or setting `make_plots=False`."
        )
    elif make_plots and use_plt:
        fig = plt.figure()
        ax = fig.gca()
        if len(errors) >= 2:
            ax.plot(
                overviews[: len(errors)],
                np.gradient(np.array(errors)),
                marker="*",
                color="tab:red",
                linestyle=":",
                label="gradient of error",
            )
        ax.plot(
            overviews[: len(errors)],
            np.array(errors),
            marker="o",
            color="tab:red",
            label="error",
        )
        ax.axhline(
            y=max_error, color="k", linestyle=":", label=f"max error = {max_error}"
        )
        ax.set_xlim([-2, max(overviews) + 1])
        ax.grid()
        ax.set_facecolor("lightgray")
        ax.set_xlabel("overview [-]")
        ax.set_ylabel("error [-]")
        ax.set_title(
            f"FILE = {os.path.split(shape_fh)[-1]} \n\n SELECTED OVERVIEW = {overview} \n\n GEOT = {info['geoTransform']}"
        )
        fig.legend()

        if os.path.isdir(make_plots):
            plot_fh = os.path.join(
                make_plots,
                f"{overview}_{os.path.split(shape_fh)[-1].replace('.geojson', '')}.png",
            )
            fig.savefig(plot_fh)
        plt.close(fig)
    else:
        ...

    return overview


if __name__ == "__main__":
    ...



import os
import pathlib

import pandas as pd
import pytest

import wapordl
import wapordl.toolbox.ts_queries
from wapordl.variable_descriptions import AGERA5_VARS, WAPOR3_VARS

module_path = wapordl.__path__[0]
assert "conda" not in module_path

test_data_folder = pathlib.Path(module_path).parent / "test_data"

wapor_gsutiuris = [x["gsutiuri"] for var, x in WAPOR3_VARS.items() if "L3" not in var]


@pytest.mark.parametrize("gsutiuri", wapor_gsutiuris)
def test_base_wapor(gsutiuri):
    json_string = '{ "type": "Polygon", "coordinates": [ [ [ 39.1751869, 8.9148245 ], [ 39.1749088, 8.3098793 ], [ 40.0254231, 8.3085969 ], [ 40.0270531, 8.9134473 ], [ 39.1751869, 8.9148245 ] ] ] }'
    period = ["2021-01-12", "2021-01-12"]
    data = wapordl.toolbox.ts_queries.request_ts(json_string, period, gsutiuri)
    assert len(data) > 0
    assert not isinstance(data[-1]["max"], type(None))


agera5_gsutiuris = [x["gsutiuri"] for x in AGERA5_VARS.values()]


@pytest.mark.parametrize("gsutiuri", agera5_gsutiuris)
def test_base_agera5(gsutiuri):
    json_string = '{ "type": "Polygon", "coordinates": [ [ [ 39.1751869, 8.9148245 ], [ 39.1749088, 8.3098793 ], [ 40.0254231, 8.3085969 ], [ 40.0270531, 8.9134473 ], [ 39.1751869, 8.9148245 ] ] ] }'
    period = ["2021-01-12", "2021-01-12"]
    data = wapordl.toolbox.ts_queries.request_ts(json_string, period, gsutiuri)
    assert len(data) > 0
    assert not isinstance(data[-1]["max"], type(None))


def test_all_touched():
    json_string = '{ "type": "Polygon", "coordinates": [ [ [ 39.1751869, 8.9148245 ], [ 39.1749088, 8.3098793 ], [ 40.0254231, 8.3085969 ], [ 40.0270531, 8.9134473 ], [ 39.1751869, 8.9148245 ] ] ] }'
    period = ["2021-01-12", "2021-03-25"]
    gsutiuri = "gs://fao-gismgr-wapor-3-data/DATA/WAPOR-3/MAPSET/L2-AETI-D"
    extra_post_args = {"all_touched": True}
    data_all_touched = wapordl.toolbox.ts_queries.request_ts(
        json_string, period, gsutiuri, extra_post_args=extra_post_args
    )
    extra_post_args = {"all_touched": False}
    data_not_all_touched = wapordl.toolbox.ts_queries.request_ts(
        json_string, period, gsutiuri, extra_post_args=extra_post_args
    )
    assert data_all_touched[-1]["maskedCount"] > data_not_all_touched[-1]["maskedCount"]


def test_level3():
    variable = "L3-AETI-D"
    region = "/Users/hmcoerver/Library/Mobile Documents/com~apple~CloudDocs/GitHub/wapordl/wapordl/test_data/test_GEZ.geojson"
    json_string = '{ "type": "MultiPolygon", "coordinates": [ [ [ [ 33.321400177824032, 14.260368111392115 ], [ 33.371790739277863, 14.250535806718197 ], [ 33.423410338815934, 14.133777188715419 ], [ 33.334919596750666, 14.068638170250711 ], [ 33.204641559821255, 14.165732178905653 ], [ 33.321400177824032, 14.260368111392115 ] ] ] ] }'
    scale_factor = 0.1
    period = ["2021-01-12", "2021-03-25"]
    gsutiuri = "gs://fao-gismgr-wapor-3-data/DATA/WAPOR-3/MOSAICSET/L3-AETI-D.GEZ"
    data = wapordl.toolbox.ts_queries.request_ts(json_string, period, gsutiuri)
    df1 = wapordl.toolbox.ts_queries.parse_data(data, scale_factor)
    assert df1.iloc[-1]["mean"] == 1.944007436363251

    df2 = wapordl.wapor_ts(region, variable, period, method="local")
    assert df1["mean"].corr(df2["mean"]) > 0.999

    gsutiuri = "gs://fao-gismgr-wapor-3-data/DATA/WAPOR-3/MOSAICSET/L3-AETI-D.AWA"
    data = wapordl.toolbox.ts_queries.request_ts(json_string, period, gsutiuri)
    df3 = wapordl.toolbox.ts_queries.parse_data(data, scale_factor)
    assert df3["mean"].isna().all()


cases = [
    ([33.2, 14.0, 33.4, 14.2], "L3-AETI-D", "none", None),
    ([33.2, 14.0, 33.4, 14.2], "L2-AETI-D", "none", None),
    ([33.2, 14.0, 33.4, 14.2], "L1-AETI-D", "none", None),
    (
        os.path.join(test_data_folder, "polygon_EPSG32636.gpkg"),
        "L1-AETI-D",
        "none",
        "name",
    ),
    (
        os.path.join(test_data_folder, "polygon_EPSG32636.gpkg"),
        "L3-AETI-D",
        "dekad",
        "bla",
    ),
    (
        os.path.join(test_data_folder, "polygon_EPSG32636.gpkg"),
        "L2-AETI-M",
        "day",
        "bla",
    ),
    (
        os.path.join(test_data_folder, "polygon_EPSG32636.gpkg"),
        "L1-AETI-D",
        "none",
        None,
    ),
    (
        os.path.join(test_data_folder, "polygon_EPSG32636.gpkg"),
        "AGERA5-ET0-E",
        "none",
        None,
    ),
]


@pytest.mark.parametrize("region,variable,unit_conversion,identifier", cases)
def test_wapor_ts(region, variable, unit_conversion, identifier):
    if variable[-2:] == "-E":
        period = ["2021-01-22", "2021-02-03"]
    else:
        period = ["2021-01-12", "2021-02-25"]

    df1 = wapordl.wapor_ts(
        region,
        variable,
        period,
        method="local",
        unit_conversion=unit_conversion,
        identifier=identifier,
    )
    df2 = wapordl.wapor_ts(
        region,
        variable,
        period,
        method="cloud",
        unit_conversion=unit_conversion,
        identifier=identifier,
    )

    if identifier in df1.columns:
        for name in df1[identifier].unique():
            if isinstance(name, type(pd.NA)):
                x1 = df1.loc[df1[identifier].isnull()]["mean"]
                x2 = df2.loc[df2[identifier].isnull()]["mean"]
            else:
                x1 = df1.loc[df1[identifier] == name]["mean"]
                x2 = df2.loc[df2[identifier] == name]["mean"]
            r = x1.corr(x2)
            print(name, r)
            assert r > 0.999
            assert df1.attrs == df2.attrs
    else:
        x1 = df1["mean"]
        x2 = df2["mean"]
        r = x1.corr(x2)
        # print(name, r)
        assert r > 0.999
        assert df1.attrs == df2.attrs


if __name__ == "__main__":
    region, variable, unit_conversion, identifier = cases[-1]

    # import matplotlib.pyplot as plt

    # fig = plt.figure(1)
    # ax = fig.gca()
    # df2.loc[df2["name"] == "corn"]["mean"].plot(ax=ax, label="new")
    # df1.loc[df1["name"] == "corn"]["mean"].plot(ax=ax, label="original")
    # plt.legend()



import logging

import requests

from wapordl.main import (
    wapor_dl,
    wapor_map,
    wapor_ts,
)

from wapordl import variable_descriptions, region_selector

__all__ = [
    "wapor_dl",
    "wapor_map",
    "wapor_ts",
    "variable_descriptions",
    "region_selector",
]
__version__ = "1.2.1"

logging.basicConfig(
    encoding="utf-8",
    level=logging.INFO,
    format="%(levelname)s: %(message)s",
    force=True,
)


def check_pywapor_version():
    """Check if the current version string is equal to
    the latest version available on pypi. If not, gives a
    warning.
    """
    package = "wapordl"
    logging.info(f"WaPORDL (`{__version__}`)")
    try:
        response = requests.get(f"https://pypi.org/pypi/{package}/json")
        response.raise_for_status()
    except Exception as _:
        ...
    else:
        latest_version = response.json()["info"]["version"]
        if latest_version != __version__:
            # ...
            logging.warning(f"Latest version is '{latest_version}'.")
            logging.warning("Please update pywapor.")


check_pywapor_version()


import os

import pandas as pd

from wapordl.toolbox.query import collect_responses


def collect_metadata(variable: str) -> dict:
    """Queries `long_name`, `units` and `source` for a given WaPOR variable code.

    Parameters
    ----------
    variable : str
        Name of variable, e.g. `L3-AETI-D`.

    Returns
    -------
    dict
        Metadata for the variable.

    Raises
    ------
    ValueError
        No valid variable name given.
    """

    if variable in AGERA5_VARS.keys():
        return {
            k: v
            for k, v in AGERA5_VARS[variable].items()
            if k in ["long_name", "units"]
        }

    if variable in WAPOR3_VARS.keys():
        return {
            k: v
            for k, v in WAPOR3_VARS[variable].items()
            if k in ["long_name", "units"]
        }

    if "L1" in variable:
        base_url = (
            "https://data.apps.fao.org/gismgr/api/v2/catalog/workspaces/WAPOR-3/mapsets"
        )
    elif "L2" in variable:
        base_url = (
            "https://data.apps.fao.org/gismgr/api/v2/catalog/workspaces/WAPOR-3/mapsets"
        )
    elif "L3" in variable:
        base_url = "https://data.apps.fao.org/gismgr/api/v2/catalog/workspaces/WAPOR-3/mosaicsets"
    elif "AGERA5" in variable:
        base_url = (
            "https://data.apps.fao.org/gismgr/api/v2/catalog/workspaces/C3S/mapsets"
        )
    else:
        raise ValueError(f"Invalid variable name {variable}.")  # NOTE: TESTED
    info = ["code", "measureCaption", "measureUnit"]
    var_codes = {
        x[0]: {"long_name": x[1], "units": x[2]}
        for x in collect_responses(base_url, info=info)
    }

    return var_codes[variable]


def date_func(url: str, tres: str) -> dict:
    """Determines start and end dates from a string a given temporal resolution, as well
    as the number of days between the two dates.

    Parameters
    ----------
    url : str
        URL linking to a resource.
    tres : str
        One of `"E"` (daily), `"D"` (dekadal), `"M"` (monthly), `"A"` (annual).

    Returns
    -------
    dict
        Dates and related information for a resource URL.

    Raises
    ------
    ValueError
        No valid `tres` given.
    """
    if tres == "D":
        year, month, dekad = os.path.split(url)[-1].split(".")[-2].split("-")
        start_day = {
            "D1": "01",
            "D2": "11",
            "D3": "21",
            "1": "01",
            "2": "11",
            "3": "21",
        }[dekad]
        start_date = f"{year}-{month}-{start_day}"
        end_day = {
            "D1": "10",
            "D2": "20",
            "D3": pd.Timestamp(start_date).daysinmonth,
            "1": "10",
            "2": "20",
            "3": pd.Timestamp(start_date).daysinmonth,
        }[dekad]
        end_date = f"{year}-{month}-{end_day}"
    elif tres == "M":
        year, month = os.path.split(url)[-1].split(".")[-2].split("-")
        start_date = f"{year}-{month}-01"
        end_date = f"{year}-{month}-{pd.Timestamp(start_date).days_in_month}"
    elif tres == "A":
        year = os.path.split(url)[-1].split(".")[-2]
        start_date = f"{year}-01-01"
        end_date = f"{year}-12-31"
    elif tres == "E":
        year, month, start_day = os.path.split(url)[-1].split(".")[-2].split("-")
        start_date = end_date = f"{year}-{month}-{start_day}"
    else:
        raise ValueError("Invalid temporal resolution.")  # NOTE: TESTED

    number_of_days = (
        pd.Timestamp(end_date) - pd.Timestamp(start_date) + pd.Timedelta(1, "D")
    ).days

    date_md = {
        "start_date": start_date,
        "end_date": end_date,
        "number_of_days": number_of_days,
        "temporal_resolution": {"E": "Day", "D": "Dekad", "M": "Month", "A": "Year"}[
            tres
        ],
    }

    if tres == "E":
        dekad = min(3, ((int(start_day) - 1) // 10) + 1)
        days_in_dekad = {1: 10, 2: 10, 3: pd.Timestamp(start_date).daysinmonth - 20}[
            dekad
        ]
        date_md["days_in_dekad"] = days_in_dekad

    return date_md


AGERA5_VARS = {
    "AGERA5-ET0-E": {
        "long_name": "Reference Evapotranspiration",
        "units": "mm/day",
        "source": "FAO56 with agERA5",
        "scale": 1.0,
        "gsutiuri": "gs://fao-gismgr-c3s-data/DATA/C3S/MAPSET/AGERA5-ET0",
    },
    "AGERA5-ET0-D": {
        "long_name": "Reference Evapotranspiration",
        "units": "mm/dekad",
        "source": "FAO56 with agERA5",
        "scale": 1.0,
        "gsutiuri": "gs://fao-gismgr-c3s-data/DATA/C3S/MAPSET/AGERA5-ET0-D",
    },
    "AGERA5-ET0-M": {
        "long_name": "Reference Evapotranspiration",
        "units": "mm/month",
        "source": "FAO56 with agERA5",
        "scale": 1.0,
        "gsutiuri": "gs://fao-gismgr-c3s-data/DATA/C3S/MAPSET/AGERA5-ET0-M",
    },
    "AGERA5-ET0-A": {
        "long_name": "Reference Evapotranspiration",
        "units": "mm/year",
        "source": "FAO56 with agERA5",
        "scale": 1.0,
        "gsutiuri": "gs://fao-gismgr-c3s-data/DATA/C3S/MAPSET/AGERA5-ET0-A",
    },
    "AGERA5-TMIN-E": {
        "long_name": "Minimum Air Temperature (2m)",
        "units": "K",
        "source": "agERA5",
        "scale": 1.0,
        "gsutiuri": "gs://fao-gismgr-c3s-data/DATA/C3S/MAPSET/AGERA5-TMIN",
    },
    "AGERA5-TMAX-E": {
        "long_name": "Maximum Air Temperature (2m)",
        "units": "K",
        "source": "agERA5",
        "scale": 1.0,
        "gsutiuri": "gs://fao-gismgr-c3s-data/DATA/C3S/MAPSET/AGERA5-TMAX",
    },
    "AGERA5-SRF-E": {
        "long_name": "Solar Radiation",
        "units": "J/m2/day",
        "source": "agERA5",
        "scale": 1.0,
        "gsutiuri": "gs://fao-gismgr-c3s-data/DATA/C3S/MAPSET/AGERA5-SRF",
    },
    "AGERA5-WS-E": {
        "long_name": "Wind Speed",
        "units": "m/s",
        "source": "agERA5",
        "scale": 1.0,
        "gsutiuri": "gs://fao-gismgr-c3s-data/DATA/C3S/MAPSET/AGERA5-WS",
    },
    "AGERA5-RH06-E": {
        "long_name": "Relative humidity at 06h (local time, 2m)",
        "units": "%",
        "source": "agERA5",
        "scale": 1.0,
        "gsutiuri": "gs://fao-gismgr-c3s-data/DATA/C3S/MAPSET/AGERA5-RH06",
    },
    "AGERA5-RH09-E": {
        "long_name": "Relative humidity at 09h (local time, 2m)",
        "units": "%",
        "source": "agERA5",
        "scale": 1.0,
        "gsutiuri": "gs://fao-gismgr-c3s-data/DATA/C3S/MAPSET/AGERA5-RH09",
    },
    "AGERA5-RH12-E": {
        "long_name": "Relative humidity at 12h (local time, 2m)",
        "units": "%",
        "source": "agERA5",
        "scale": 1.0,
        "gsutiuri": "gs://fao-gismgr-c3s-data/DATA/C3S/MAPSET/AGERA5-RH12",
    },
    "AGERA5-RH15-E": {
        "long_name": "Relative humidity at 15h (local time, 2m)",
        "units": "%",
        "source": "agERA5",
        "scale": 1.0,
        "gsutiuri": "gs://fao-gismgr-c3s-data/DATA/C3S/MAPSET/AGERA5-RH15",
    },
    "AGERA5-RH18-E": {
        "long_name": "Relative humidity at 18h (local time, 2m)",
        "units": "%",
        "source": "agERA5",
        "scale": 1.0,
        "gsutiuri": "gs://fao-gismgr-c3s-data/DATA/C3S/MAPSET/AGERA5-RH18",
    },
    "AGERA5-PF-E": {
        "long_name": "Precipitation",
        "units": "mm/day",
        "source": "agERA5",
        "scale": 1.0,
        "gsutiuri": "gs://fao-gismgr-c3s-data/DATA/C3S/MAPSET/AGERA5-PF",
    },
    "AGERA5-PF-D": {
        "long_name": "Precipitation",
        "units": "mm/dekad",
        "source": "agERA5",
        "scale": 1.0,
        "gsutiuri": "gs://fao-gismgr-c3s-data/DATA/C3S/MAPSET/AGERA5-PF-D",
    },
    "AGERA5-PF-M": {
        "long_name": "Precipitation",
        "units": "mm/month",
        "source": "agERA5",
        "scale": 1.0,
        "gsutiuri": "gs://fao-gismgr-c3s-data/DATA/C3S/MAPSET/AGERA5-PF-M",
    },
    "AGERA5-PF-A": {
        "long_name": "Precipitation",
        "units": "mm/year",
        "source": "agERA5",
        "scale": 1.0,
        "gsutiuri": "gs://fao-gismgr-c3s-data/DATA/C3S/MAPSET/AGERA5-PF-A",
    },
}

# new = dict()
# from wapordl.products.wapor3 import generate_urls
# period = ["2021-01-01", "2021-02-01"]

# for variable, meta in AGERA5_VARS.items():
#     if variable[-2:] == "-E":
#         variable_ = variable[:-2]
#     if variable in new.keys():
#         continue
#     print(variable)
#     x = generate_urls(variable_, None, period, info="scale")
#     y = generate_urls(variable_, None, period, info="gsutilUri")
#     new[variable] = {**meta, **{"scale": x[0]}, **{"gsutiuri": y[0]}}

WAPOR3_VARS = {
    "L1-AETI-A": {
        "long_name": "Actual EvapoTranspiration and Interception",
        "units": "mm/year",
        "gsutiuri": "gs://fao-gismgr-wapor-3-data/DATA/WAPOR-3/MAPSET/L1-AETI-A",
        "scale": 0.1,
    },
    "L1-AETI-D": {
        "long_name": "Actual EvapoTranspiration and Interception",
        "units": "mm/day",
        "gsutiuri": "gs://fao-gismgr-wapor-3-data/DATA/WAPOR-3/MAPSET/L1-AETI-D",
        "scale": 0.1,
    },
    "L1-AETI-M": {
        "long_name": "Actual EvapoTranspiration and Interception",
        "units": "mm/month",
        "gsutiuri": "gs://fao-gismgr-wapor-3-data/DATA/WAPOR-3/MAPSET/L1-AETI-M",
        "scale": 0.1,
    },
    "L1-E-A": {
        "long_name": "Evaporation",
        "units": "mm/year",
        "gsutiuri": "gs://fao-gismgr-wapor-3-data/DATA/WAPOR-3/MAPSET/L1-E-A",
        "scale": 0.1,
    },
    "L1-E-D": {
        "long_name": "Evaporation",
        "units": "mm/day",
        "gsutiuri": "gs://fao-gismgr-wapor-3-data/DATA/WAPOR-3/MAPSET/L1-E-D",
        "scale": 0.1,
    },
    "L1-GBWP-A": {
        "long_name": "Gross Biomass Water Productivity",
        "units": "kg/m³",
        "gsutiuri": "gs://fao-gismgr-wapor-3-data/DATA/WAPOR-3/MAPSET/L1-GBWP-A",
        "scale": 0.001,
    },
    "L1-I-A": {
        "long_name": "Interception",
        "units": "mm/year",
        "gsutiuri": "gs://fao-gismgr-wapor-3-data/DATA/WAPOR-3/MAPSET/L1-I-A",
        "scale": 0.1,
    },
    "L1-I-D": {
        "long_name": "Interception",
        "units": "mm/day",
        "gsutiuri": "gs://fao-gismgr-wapor-3-data/DATA/WAPOR-3/MAPSET/L1-I-D",
        "scale": 0.1,
    },
    "L1-NBWP-A": {
        "long_name": "Net Biomass Water Productivity",
        "units": "kg/m³",
        "gsutiuri": "gs://fao-gismgr-wapor-3-data/DATA/WAPOR-3/MAPSET/L1-NBWP-A",
        "scale": 0.001,
    },
    "L1-NPP-D": {
        "long_name": "Net Primary Production",
        "units": "gC/m²/day",
        "gsutiuri": "gs://fao-gismgr-wapor-3-data/DATA/WAPOR-3/MAPSET/L1-NPP-D",
        "scale": 0.001,
    },
    "L1-NPP-M": {
        "long_name": "Net Primary Production",
        "units": "gC/m²/month",
        "gsutiuri": "gs://fao-gismgr-wapor-3-data/DATA/WAPOR-3/MAPSET/L1-NPP-M",
        "scale": 0.001,
    },
    "L1-PCP-A": {
        "long_name": "Precipitation",
        "units": "mm/year",
        "gsutiuri": "gs://fao-gismgr-wapor-3-data/DATA/WAPOR-3/MAPSET/L1-PCP-A",
        "scale": 0.1,
    },
    "L1-PCP-D": {
        "long_name": "Precipitation",
        "units": "mm/day",
        "gsutiuri": "gs://fao-gismgr-wapor-3-data/DATA/WAPOR-3/MAPSET/L1-PCP-D",
        "scale": 0.1,
    },
    "L1-PCP-E": {
        "long_name": "Precipitation",
        "units": "mm/day",
        "gsutiuri": "gs://fao-gismgr-wapor-3-data/DATA/WAPOR-3/MAPSET/L1-PCP-E",
        "scale": 0.1,
    },
    "L1-PCP-M": {
        "long_name": "Precipitation",
        "units": "mm/month",
        "gsutiuri": "gs://fao-gismgr-wapor-3-data/DATA/WAPOR-3/MAPSET/L1-PCP-M",
        "scale": 0.1,
    },
    "L1-QUAL-LST-D": {
        "long_name": "Quality Land Surface Temperature",
        "units": "d",
        "gsutiuri": "gs://fao-gismgr-wapor-3-data/DATA/WAPOR-3/MAPSET/L1-QUAL-LST-D",
        "scale": 1.0,
    },
    "L1-QUAL-NDVI-D": {
        "long_name": "Quality of Normalized Difference Vegetation Index",
        "units": "d",
        "gsutiuri": "gs://fao-gismgr-wapor-3-data/DATA/WAPOR-3/MAPSET/L1-QUAL-NDVI-D",
        "scale": 1.0,
    },
    "L1-RET-A": {
        "long_name": "Reference Evapotranspiration",
        "units": "mm/year",
        "gsutiuri": "gs://fao-gismgr-wapor-3-data/DATA/WAPOR-3/MAPSET/L1-RET-A",
        "scale": 0.1,
    },
    "L1-RET-D": {
        "long_name": "Reference Evapotranspiration",
        "units": "mm/day",
        "gsutiuri": "gs://fao-gismgr-wapor-3-data/DATA/WAPOR-3/MAPSET/L1-RET-D",
        "scale": 0.1,
    },
    "L1-RET-E": {
        "long_name": "Reference Evapotranspiration",
        "units": "mm/day",
        "gsutiuri": "gs://fao-gismgr-wapor-3-data/DATA/WAPOR-3/MAPSET/L1-RET-E",
        "scale": 0.1,
    },
    "L1-RET-M": {
        "long_name": "Reference Evapotranspiration",
        "units": "mm/month",
        "gsutiuri": "gs://fao-gismgr-wapor-3-data/DATA/WAPOR-3/MAPSET/L1-RET-M",
        "scale": 0.1,
    },
    "L1-RSM-D": {
        "long_name": "Relative Soil Moisture",
        "units": "%",
        "gsutiuri": "gs://fao-gismgr-wapor-3-data/DATA/WAPOR-3/MAPSET/L1-RSM-D",
        "scale": 0.001,
    },
    "L1-T-A": {
        "long_name": "Transpiration",
        "units": "mm/year",
        "gsutiuri": "gs://fao-gismgr-wapor-3-data/DATA/WAPOR-3/MAPSET/L1-T-A",
        "scale": 0.1,
    },
    "L1-T-D": {
        "long_name": "Transpiration",
        "units": "mm/day",
        "gsutiuri": "gs://fao-gismgr-wapor-3-data/DATA/WAPOR-3/MAPSET/L1-T-D",
        "scale": 0.1,
    },
    "L1-TBP-A": {
        "long_name": "Total Biomass Production",
        "units": "kg/ha",
        "gsutiuri": "gs://fao-gismgr-wapor-3-data/DATA/WAPOR-3/MAPSET/L1-TBP-A",
        "scale": 1.0,
    },
    "L2-AETI-A": {
        "long_name": "Actual EvapoTranspiration and Interception",
        "units": "mm/year",
        "gsutiuri": "gs://fao-gismgr-wapor-3-data/DATA/WAPOR-3/MAPSET/L2-AETI-A",
        "scale": 0.1,
    },
    "L2-AETI-D": {
        "long_name": "Actual EvapoTranspiration and Interception",
        "units": "mm/day",
        "gsutiuri": "gs://fao-gismgr-wapor-3-data/DATA/WAPOR-3/MAPSET/L2-AETI-D",
        "scale": 0.1,
    },
    "L2-AETI-M": {
        "long_name": "Actual EvapoTranspiration and Interception",
        "units": "mm/month",
        "gsutiuri": "gs://fao-gismgr-wapor-3-data/DATA/WAPOR-3/MAPSET/L2-AETI-M",
        "scale": 0.1,
    },
    "L2-E-A": {
        "long_name": "Evaporation",
        "units": "mm/year",
        "gsutiuri": "gs://fao-gismgr-wapor-3-data/DATA/WAPOR-3/MAPSET/L2-E-A",
        "scale": 0.1,
    },
    "L2-E-D": {
        "long_name": "Evaporation",
        "units": "mm/day",
        "gsutiuri": "gs://fao-gismgr-wapor-3-data/DATA/WAPOR-3/MAPSET/L2-E-D",
        "scale": 0.1,
    },
    "L2-GBWP-A": {
        "long_name": "Gross Biomass Water Productivity",
        "units": "kg/m³",
        "gsutiuri": "gs://fao-gismgr-wapor-3-data/DATA/WAPOR-3/MAPSET/L2-GBWP-A",
        "scale": 0.001,
    },
    "L2-I-A": {
        "long_name": "Interception",
        "units": "mm/year",
        "gsutiuri": "gs://fao-gismgr-wapor-3-data/DATA/WAPOR-3/MAPSET/L2-I-A",
        "scale": 0.1,
    },
    "L2-I-D": {
        "long_name": "Interception",
        "units": "mm/day",
        "gsutiuri": "gs://fao-gismgr-wapor-3-data/DATA/WAPOR-3/MAPSET/L2-I-D",
        "scale": 0.1,
    },
    "L2-NBWP-A": {
        "long_name": "Net Biomass Water Productivity",
        "units": "kg/m³",
        "gsutiuri": "gs://fao-gismgr-wapor-3-data/DATA/WAPOR-3/MAPSET/L2-NBWP-A",
        "scale": 0.001,
    },
    "L2-NPP-D": {
        "long_name": "Net Primary Production",
        "units": "gC/m²/day",
        "gsutiuri": "gs://fao-gismgr-wapor-3-data/DATA/WAPOR-3/MAPSET/L2-NPP-D",
        "scale": 0.001,
    },
    "L2-NPP-M": {
        "long_name": "Net Primary Production",
        "units": "gC/m²/month",
        "gsutiuri": "gs://fao-gismgr-wapor-3-data/DATA/WAPOR-3/MAPSET/L2-NPP-M",
        "scale": 0.001,
    },
    "L2-QUAL-NDVI-D": {
        "long_name": "Quality of Normalized Difference Vegetation Index",
        "units": "d",
        "gsutiuri": "gs://fao-gismgr-wapor-3-data/DATA/WAPOR-3/MAPSET/L2-QUAL-NDVI-D",
        "scale": 1.0,
    },
    "L2-RSM-D": {
        "long_name": "Relative Soil Moisture",
        "units": "%",
        "gsutiuri": "gs://fao-gismgr-wapor-3-data/DATA/WAPOR-3/MAPSET/L2-RSM-D",
        "scale": 0.001,
    },
    "L2-T-A": {
        "long_name": "Transpiration",
        "units": "mm/year",
        "gsutiuri": "gs://fao-gismgr-wapor-3-data/DATA/WAPOR-3/MAPSET/L2-T-A",
        "scale": 0.1,
    },
    "L2-T-D": {
        "long_name": "Transpiration",
        "units": "mm/day",
        "gsutiuri": "gs://fao-gismgr-wapor-3-data/DATA/WAPOR-3/MAPSET/L2-T-D",
        "scale": 0.1,
    },
    "L2-TBP-A": {
        "long_name": "Total Biomass Production",
        "units": "kg/ha",
        "gsutiuri": "gs://fao-gismgr-wapor-3-data/DATA/WAPOR-3/MAPSET/L2-TBP-A",
        "scale": 1.0,
    },
    "L3-AETI-A": {
        "long_name": "Actual EvapoTranspiration and Interception",
        "units": "mm/year",
        "gsutiuri": "gs://fao-gismgr-wapor-3-data/DATA/WAPOR-3/MOSAICSET/L3-AETI-A",
        "scale": 0.1,
    },
    "L3-AETI-D": {
        "long_name": "Actual EvapoTranspiration and Interception",
        "units": "mm/day",
        "gsutiuri": "gs://fao-gismgr-wapor-3-data/DATA/WAPOR-3/MOSAICSET/L3-AETI-D",
        "scale": 0.1,
    },
    "L3-AETI-M": {
        "long_name": "Actual EvapoTranspiration and Interception",
        "units": "mm/month",
        "gsutiuri": "gs://fao-gismgr-wapor-3-data/DATA/WAPOR-3/MOSAICSET/L3-AETI-M",
        "scale": 0.1,
    },
    "L3-E-A": {
        "long_name": "Evaporation",
        "units": "mm/year",
        "gsutiuri": "gs://fao-gismgr-wapor-3-data/DATA/WAPOR-3/MOSAICSET/L3-E-A",
        "scale": 0.1,
    },
    "L3-E-D": {
        "long_name": "Evaporation",
        "units": "mm/day",
        "gsutiuri": "gs://fao-gismgr-wapor-3-data/DATA/WAPOR-3/MOSAICSET/L3-E-D",
        "scale": 0.1,
    },
    "L3-GBWP-A": {
        "long_name": "Gross Biomass Water Productivity",
        "units": "kg/m³",
        "gsutiuri": "gs://fao-gismgr-wapor-3-data/DATA/WAPOR-3/MOSAICSET/L3-GBWP-A",
        "scale": 0.001,
    },
    "L3-I-A": {
        "long_name": "Interception",
        "units": "mm/year",
        "gsutiuri": "gs://fao-gismgr-wapor-3-data/DATA/WAPOR-3/MOSAICSET/L3-I-A",
        "scale": 0.1,
    },
    "L3-I-D": {
        "long_name": "Interception",
        "units": "mm/day",
        "gsutiuri": "gs://fao-gismgr-wapor-3-data/DATA/WAPOR-3/MOSAICSET/L3-I-D",
        "scale": 0.1,
    },
    "L3-NBWP-A": {
        "long_name": "Net Biomass Water Productivity",
        "units": " kg/m³",
        "gsutiuri": "gs://fao-gismgr-wapor-3-data/DATA/WAPOR-3/MOSAICSET/L3-NBWP-A",
        "scale": 0.001,
    },
    "L3-NPP-D": {
        "long_name": "Net Primary Production",
        "units": "gC/m²/day",
        "gsutiuri": "gs://fao-gismgr-wapor-3-data/DATA/WAPOR-3/MOSAICSET/L3-NPP-D",
        "scale": 0.001,
    },
    "L3-NPP-M": {
        "long_name": "Net Primary Production",
        "units": "gC/m²/month",
        "gsutiuri": "gs://fao-gismgr-wapor-3-data/DATA/WAPOR-3/MOSAICSET/L3-NPP-M",
        "scale": 0.001,
    },
    "L3-QUAL-NDVI-D": {
        "long_name": "Quality of Normalized Difference Vegetation Index",
        "units": "d",
        "gsutiuri": "gs://fao-gismgr-wapor-3-data/DATA/WAPOR-3/MOSAICSET/L3-QUAL-NDVI-D",
        "scale": 1.0,
    },
    "L3-RSM-D": {
        "long_name": "Relative Soil Moisture",
        "units": "%",
        "gsutiuri": "gs://fao-gismgr-wapor-3-data/DATA/WAPOR-3/MOSAICSET/L3-RSM-D",
        "scale": 0.001,
    },
    "L3-T-A": {
        "long_name": "Transpiration",
        "units": "mm/year",
        "gsutiuri": "gs://fao-gismgr-wapor-3-data/DATA/WAPOR-3/MOSAICSET/L3-T-A",
        "scale": 0.1,
    },
    "L3-T-D": {
        "long_name": "Transpiration",
        "units": "mm/day",
        "gsutiuri": "gs://fao-gismgr-wapor-3-data/DATA/WAPOR-3/MOSAICSET/L3-T-D",
        "scale": 0.1,
    },
    "L3-TBP-A": {
        "long_name": "Total Biomass Production",
        "units": "kg/ha",
        "gsutiuri": "gs://fao-gismgr-wapor-3-data/DATA/WAPOR-3/MOSAICSET/L3-TBP-A",
        "scale": 1.0,
    },
}



import importlib.util
import logging
import os
import time
from typing import List, Literal, Tuple

import numpy as np
import pandas as pd
from osgeo import gdal, gdalconst, ogr, osr

from wapordl.unit_convertor import unit_convertor

optional_packages = ["tqdm"]
use_tqdm = all(
    [not isinstance(importlib.util.find_spec(x), type(None)) for x in optional_packages]
)
if use_tqdm:
    from tqdm import tqdm

gdal.UseExceptions()


def extension_in_gdal_drivers(extension, raise_error=False):
    succes = False
    for i in range(gdal.GetDriverCount()):
        drv = gdal.GetDriver(i)
        extensions = drv.GetMetadataItem(gdal.DMD_EXTENSIONS)
        # md = drv.GetMetadata_Dict()
        # shortname = drv.ShortName
        # longname = drv.LongName
        # is_raster = 'DCAP_RASTER' in md
        # is_vector = 'DCAP_VECTOR' in md
        if extensions is not None:
            if extension.replace(".", "") in extensions:
                succes = True
    guess = {
        ".nc": ("libgdal-netcdf", "netCDF"),
        ".jp2": ("libgdal-jp2openjpeg", "JP2OpenJPEG"),
    }
    if not succes:
        if extension in guess.keys():
            logging.warning(f"No driver found for `{guess[extension][1]}`.")
            logging.warning(
                f"Run `conda install -c conda-forge {guess[extension][0]}` to install a driver for `{guess[extension][1]}`."
            )
        else:
            logging.warning(f"No driver found for `{extension}`.")
        if raise_error:
            raise ValueError(
                f"No driver found for `{guess.get(extension, ['', extension])[1]}`."
            )
    return succes


_ = extension_in_gdal_drivers(".tif")
_ = extension_in_gdal_drivers(".nc")
_ = extension_in_gdal_drivers(".shp")
_ = extension_in_gdal_drivers(".geojson")
_ = extension_in_gdal_drivers(".gpkg")


def unlink_vsimems(paths):
    if not isinstance(paths, list):
        paths = [paths]
    for path in paths:
        if isinstance(path, str):
            if "/vsimem/" in path:
                try:
                    _ = gdal.Unlink(path)
                except RuntimeError:
                    ...


############
## RASTER ##
############


def cog_dl(
    urls: List[str],
    out_fn: str,
    overview: str | int = "NONE",
    warp_kwargs: dict = {},
    vrt_options: dict = {"separate": True},
    unit_conversion: str = "none",
) -> Tuple[str]:
    """Download multiple COGs into the bands of a single geotif or netcdf file.

    Parameters
    ----------
    urls : list
        URLs of the different COGs to be downloaded.
    out_fn : str
        Path to the output file.
    overview : str | int, optional
        Select which overview from the COGs to use. Can be "NONE" or -1 to use the original data,
        an integer from 0 up to the total of overviews available or "AUTO" for
        automatic overview detection, by default "NONE".
    warp_kwargs : dict, optional
        Additional gdal.Warp keyword arguments, by default {}.
    vrt_options : dict, optional
        Additional options passed to gdal.BuildVRT, by default {"separate": True}.
    unit_conversion : str, optional
        Apply a unit conversion on the created file, can be one of "none", "day", "dekad",
        "month" or "year", by default "none".

    Returns
    -------
    tuple
        Paths to the created geotiff file and the (intermediate) vrt file.

    Raises
    ------
    ValueError
        Invalid output extension selected.
    """
    global use_tqdm

    out_ext = os.path.splitext(out_fn)[-1]
    valid_ext = {".nc": "netCDF", ".tif": "GTiff"}
    valid_cos = {".nc": ["COMPRESS=DEFLATE", "FORMAT=NC4C"], ".tif": ["COMPRESS=LZW"]}
    if not bool(np.isin(out_ext, list(valid_ext.keys()))):
        raise ValueError(
            f"Please use one of {list(valid_ext.keys())} as extension for `out_fn`, not {out_ext}"
        )  # NOTE: TESTED
    vrt_fn = out_fn.replace(out_ext, ".vrt")

    ## Build VRT with all the required data.
    vrt_options_ = gdal.BuildVRTOptions(**vrt_options)
    prepend = {False: "/vsicurl/", True: "/vsigzip//vsicurl/"}
    vrt = gdal.BuildVRT(
        vrt_fn, [prepend[".gz" in x[1]] + x[1] for x in urls], options=vrt_options_
    )
    vrt.FlushCache()

    n_urls = len(urls)

    if use_tqdm:
        # Create waitbar.
        waitbar = tqdm(
            desc=f"Downloading {n_urls} COGs",
            leave=False,
            total=100,
            bar_format="{l_bar}{bar}|",
        )

        # Define callback function for waitbar progress.
        def _callback_func(info, *args):
            waitbar.update(info * 100 - waitbar.n)
    else:
        waitbar = _callback_func = None
        logging.info("Consider installing `tqdm` to display a download progress bar.")

    ## Download the data.
    warp_options = gdal.WarpOptions(
        format=valid_ext[out_ext],
        cropToCutline=True,
        overviewLevel=overview,
        multithread=True,
        targetAlignedPixels=True,
        creationOptions=valid_cos[out_ext],
        callback=_callback_func,
        warpOptions=["CUTLINE_ALL_TOUCHED", "TRUE"],
        **warp_kwargs,
    )
    warp = gdal.Warp(out_fn, vrt_fn, options=warp_options)
    warp.FlushCache()  # NOTE do not remove this.

    if waitbar is not None:
        waitbar.close()

    nbands = warp.RasterCount

    if nbands == n_urls and unit_conversion != "none":
        out_fn_new = out_fn.replace(out_ext, f"_converted{out_ext}")
        out_fn_old = out_fn
        warp, out_fn = unit_convertor(
            urls, out_fn, out_fn_new, unit_conversion, warp, coptions=valid_cos[out_ext]
        )
    else:
        out_fn_old = ""

    if nbands == n_urls:
        for i, (md, _) in enumerate(urls):
            if not isinstance(md, type(None)):
                band = warp.GetRasterBand(i + 1)
                band.SetDescription(md.get("start_date", f"Band {i + 1}"))
                band.SetMetadata(md)

    warp.FlushCache()

    if os.path.isfile(vrt_fn):
        try:
            os.remove(vrt_fn)
        except PermissionError:
            ...

    if os.path.isfile(out_fn_old) and os.path.isfile(out_fn_new):
        try:
            os.remove(out_fn_old)
        except PermissionError:
            ...

    return out_fn, vrt_fn


def get_stats(path: str) -> pd.DataFrame:
    """Get statistics for a raster file.

    Parameters
    ----------
    path : str
        Path to raster file.

    Returns
    -------
    pd.DataFrame
        The calculated statistics.
    """
    try:
        stats = gdal.Info(path, format="json", stats=True)
    except RuntimeError as e:
        if "Failed to compute statistics, no valid pixels found in sampling." in str(e):
            stats = gdal.Info(path, format="json", stats=False)
        else:
            raise e
    ## Get scale and offset factor.
    scale = stats["bands"][0].get("scale", 1)
    offset = stats["bands"][0].get("offset", 0)

    ## Check offset factor.
    if offset != 0:
        logging.warning("Offset factor is not zero, statistics might be wrong.")

    data = {
        statistic: [x.get(statistic, np.nan) for x in stats["bands"]]
        for statistic in ["minimum", "maximum", "mean"]
    }
    data: pd.DataFrame = pd.DataFrame(data) * scale
    data["start_date"] = [
        pd.Timestamp(x.get("metadata", {}).get("", {}).get("start_date", "nat"))
        for x in stats["bands"]
    ]
    data["end_date"] = [
        pd.Timestamp(x.get("metadata", {}).get("", {}).get("end_date", "nat"))
        for x in stats["bands"]
    ]
    data["number_of_days"] = [
        pd.Timedelta(
            float(x.get("metadata", {}).get("", {}).get("number_of_days", np.nan)),
            "days",
        )
        for x in stats["bands"]
    ]

    data.attrs = make_df_md(path)
    return data


def translate(
    fp: str, extension: str, separate_unscale: bool = False
) -> str | List[str]:
    """Change filetype of input raster.

    Parameters
    ----------
    fp : str
        Path to raster file to translate.
    extension : str
        Extension of output file, e.g. `".nc"`.
    separate_unscale : bool, optional
        Mutually exclusive with `extension=".tif"`. Creates multiple files with one band each, by default False.

    Returns
    -------
    str | List[str]
        Path to translated raster.
    """
    if extension == ".tif" and separate_unscale:
        logging.info("Splitting single GeoTIFF into multiple unscaled files.")
        ds = gdal.Open(fp)
        number_of_bands = ds.RasterCount
        fps = list()
        for band_number in range(1, number_of_bands + 1):
            band = ds.GetRasterBand(band_number)
            md = band.GetMetadata()
            options = gdal.TranslateOptions(
                unscale=True,
                outputType=gdalconst.GDT_Float64,
                bandList=[band_number],
                creationOptions=["COMPRESS=LZW"],
            )
            output_file = fp.replace(".tif", f"_{md['start_date']}.tif")
            x = gdal.Translate(output_file, fp, options=options)
            x.FlushCache()
            fps.append(output_file)
        ds.FlushCache()
        ds = None
        try:
            os.remove(fp)
        except PermissionError:
            ...
        return fps
    elif extension != ".tif":
        _ = extension_in_gdal_drivers(extension, raise_error=True)
        if separate_unscale:
            logging.warning(
                f"The `separate_unscale` option only works with `.tif` extension, not with `{extension}`."
            )
        logging.info(f"Converting from `.tif` to `{extension}`.")
        toptions = {".nc": {"creationOptions": ["COMPRESS=DEFLATE", "FORMAT=NC4C"]}}
        options = gdal.TranslateOptions(**toptions.get(extension, {}))
        new_fp = fp.replace(".tif", extension)
        ds = gdal.Translate(new_fp, fp, options=options)
        ds.FlushCache()
        try:
            os.remove(fp)
        except PermissionError:
            ...
        return new_fp
    else:
        return fp


def get_info(url: str) -> dict:
    """Get metadata from a online raster file.

    Parameters
    ----------
    url : str
        URL pointing to the raster file.

    Returns
    -------
    dict
        The metadata.
    """
    if not os.path.isfile(url):
        url = {False: "/vsicurl/", True: "/vsigzip//vsicurl/"}[".gz" in url] + url
    info = gdal.Info(url, format="json")
    return info


def get_extent(ds):
    """Return list of corner coordinates from a gdal Dataset"""
    xmin, xpixel, _, ymax, _, ypixel = ds.GetGeoTransform()
    width, height = ds.RasterXSize, ds.RasterYSize
    xmax = xmin + width * xpixel
    ymin = ymax + height * ypixel
    return (xmin, ymin), (xmax, ymax)


def reproject_coords(coords, src_srs, tgt_srs):
    """Reproject a list of x,y coordinates."""
    trans_coords = []
    transform = osr.CoordinateTransformation(src_srs, tgt_srs)
    for x, y in coords:
        x, y, z = transform.TransformPoint(x, y)
        trans_coords += [x]
        trans_coords += [y]
    return trans_coords


def get_wgs84_bounds(path):
    ds = gdal.Open(path)
    bounds = get_extent(ds)
    src_srs = osr.SpatialReference()
    src_srs.ImportFromWkt(ds.GetProjection())
    tgt_srs = osr.SpatialReference()
    tgt_srs.ImportFromEPSG(4326)
    trans_coords = reproject_coords(bounds, src_srs, tgt_srs)
    return trans_coords


############
## VECTOR ##
############


def make_union(geoms: List[ogr.Geometry], format: Literal["json", "wkt"] | None = None):
    final_geom = geoms.pop(0)
    while geoms:
        geom_ = geoms.pop(0)
        final_geom = final_geom.Union(geom_)
    if format == "json":
        return final_geom.ExportToJson()
    elif format == "wkt":
        return final_geom.ExportToWkt()
    else:
        return final_geom


def group_geoms_on_attribute(
    vector, identifier, format: Literal["json", "wkt"] | None = None
):
    # Open the vector
    vector_ds: gdal.Dataset = gdal.OpenEx(vector, gdal.OF_VECTOR)

    # Check the number of layers.
    n_layers = vector_ds.GetLayerCount()
    if n_layers > 1:
        logging.warning(
            f"`{vector_ds}` has {n_layers} layers, only the first layer will be processed."
        )

    # Open the first layer.
    layer: List[ogr.Feature] = vector_ds.GetLayer(0)

    spatial_ref: osr.SpatialReference = layer.GetSpatialRef()
    srs = spatial_ref.ExportToWkt()

    # Determine the field names.
    layer_def: ogr.FeatureDefn = layer.GetLayerDefn()
    field_names = [
        getattr(layer_def.GetFieldDefn(x), "name")
        for x in range(layer_def.GetFieldCount())
    ]

    # Check if chosen ID is valid.
    if not isinstance(identifier, type(None)) and (identifier not in field_names):
        raise ValueError(
            f"Identifier `{identifier}` not found in field names `{field_names}`."
        )

    geoms: dict[str, List] = dict()
    for ftr in layer:
        if isinstance(identifier, type(None)):
            ftr_name = pd.NA
        else:
            ftr_name = ftr.GetField(identifier)

        if isinstance(ftr_name, type(None)):
            ftr_name = pd.NA

        geom: ogr.Geometry = ftr.geometry()
        if ftr_name in geoms.keys():
            geoms[ftr_name].append(geom.Clone())
        else:
            geoms[ftr_name] = [geom.Clone()]

    unions = {
        ftr_name: make_union(geoms, format=format) for ftr_name, geoms in geoms.items()
    }

    return unions, srs


def get_geom(
    fh: str, lyr_idx: int = 0, ftr_idx: int = 0
) -> Tuple[ogr.Geometry, ogr.Feature, ogr.Layer, gdal.Dataset]:
    """_summary_

    Parameters
    ----------
    fh : str
        Path to vector file.
    lyr_idx : int, optional
        Index of layer to retrieve, by default 0.
    ftr_idx : int, optional
        Index of feature to retrieve, by default 0.

    Returns
    -------
    Tuple[ogr.Geometry, ogr.Feature, ogr.Layer, gdal.Dataset]
        The Geometry and other parts (Feature, Layer and Dataset) of the opened file.
        Make sure that the Feature stays open, otherwise the Geometry will be closed too.
    """
    ds = gdal.OpenEx(fh, gdal.OF_VECTOR)
    layer = ds.GetLayerByIndex(lyr_idx)
    for _ in range(ftr_idx + 1):
        ftr = layer.GetNextFeature()
    geom = ftr.GetGeometryRef()
    return geom, ftr, layer, ds  # NOTE need to return `ftr` to keep layer open.


def check_intersects(base_shape: str, bbs: dict) -> dict:
    """Check if a shape intersects with bounding-boxes stored in the values of a
    dictionary.

    Parameters
    ----------
    base_shape : str
        Path to vector.
    bbs : dict
        The bounding-boxes, keys are names, values are bounding-boxes.

    Returns
    -------
    dict
        Bool for every key/value pair in `bbs` indicating if the bounding-box intersects
        with `base_shape`.
    """
    geom, _ = get_geom(base_shape)[:2]
    checks = dict()
    for name, coords in bbs.items():
        wkt = coords_to_wkt_polygon(coords)
        geom_bb = ogr.CreateGeometryFromWkt(wkt)
        checks[name] = geom_bb.Intersects(geom)
    return checks


def get_area(fh: str, lyr_idx: int = 0, ftr_idx: int = 0) -> float:
    """Get the area of a geometry from a file, layer and feature index.

    Parameters
    ----------
    fh : str
        Patht to file.
    lyr_idx : int, optional
        Which layer index to use, by default 0.
    ftr_idx : int, optional
        Which feature index from the layer to use, by default 0.

    Returns
    -------
    float
        Area of the geometry.
    """
    geom, _ = get_geom(fh, lyr_idx=lyr_idx, ftr_idx=ftr_idx)[:2]
    return geom.GetArea()


def get_bounds(
    fh: str, lyr_idx: int = 0, ftr_idx: int = 0, lrbt: bool = True
) -> List[float]:
    """Get the bounds of a geometry as [left, right, bottom, top] from a file,
    layer and feature index.

    Parameters
    ----------
    fh : str
        Patht to file.
    lyr_idx : int, optional
        Which layer index to use, by default 0.
    ftr_idx : int, optional
        Which feature index from the layer to use, by default 0.

    Returns
    -------
    List[float]
        The bounds [left, right, bottom, top].
    """
    geom, _ = get_geom(fh, lyr_idx=lyr_idx, ftr_idx=ftr_idx)[:2]
    bounds = geom.GetEnvelope()
    if lrbt:  # left right bottom top
        return [bounds[0], bounds[2], bounds[1], bounds[3]]
    else:
        return bounds


def get_wkt(fh: str, lyr_idx: int = 0, ftr_idx: int = 0) -> str:
    """Get a WKT string of a geometry from a file, layer and feature index.

    Parameters
    ----------
    fh : str
        Patht to file.
    lyr_idx : int, optional
        Which layer index to use, by default 0.
    ftr_idx : int, optional
        Which feature index from the layer to use, by default 0.

    Returns
    -------
    str
        The WKT string.
    """
    geom, _ = get_geom(fh, lyr_idx=lyr_idx, ftr_idx=ftr_idx)[:2]
    return geom.ExportToWkt()


def check_vector(fh: str) -> tuple:
    """Check if a provided vector file is correctly formatted for wapordl.

    Parameters
    ----------
    fh : str
        Path to input file.

    Returns
    -------
    tuple
        Information about the input file, first value is EPSG code (int), second is
        driver name, third is True if coordinates are 2D.
    """
    geom, _, layer, ds = get_geom(fh, lyr_idx=0, ftr_idx=0)

    driver = ds.GetDriver()
    driver_name = getattr(driver, "name", getattr(driver, "LongName", None))
    is_two_d = geom.CoordinateDimension() == 2
    spatialRef = layer.GetSpatialRef()
    epsg = spatialRef.GetAuthorityCode(None)

    try:
        ds = ds.Close()
    except AttributeError:
        ds = ds.FlushCache()

    return int(epsg), driver_name, is_two_d


def reproject_vector(fh: str, epsg=4326, in_memory=False) -> str:
    """Create a 2D GeoJSON file with `EPSG:4326` SRS from any
    OGR compatible vector file.

    Parameters
    ----------
    fh : str
        Path to input file.
    epsg : int, optional
        target SRS, by default 4326.

    Returns
    -------
    str
        Path to output (GeoJSON) file.
    """

    ext = os.path.splitext(fh)[-1]
    out_fh = fh.replace(ext, f"_reprojected_{epsg}.geojson")

    if "/vsimem/" not in out_fh and in_memory:
        out_fh = f"/vsimem/x_{time.strftime('%Y-%m-%d_%H%M%S')}_{np.random.randint(9999)}.geojson"

    options = gdal.VectorTranslateOptions(
        dstSRS=f"EPSG:{epsg}",
        format="GeoJSON",
        dim="XY",
    )
    x = gdal.VectorTranslate(out_fh, fh, options=options)
    x.FlushCache()
    x = None

    return out_fh


def bb_to_coords(bb: list[float]) -> List[List[float]]:
    """Convert a bounding-box to a list of coordinates.

    Parameters
    ----------
    bb : List[float]
        Bounding-box as [xmin, ymin, xmax, ymax].

    Returns
    -------
    List[List[float]]
        The coordinates of the cornerpoints of the given bounding-box.
    """
    coords = [
        [bb[0], bb[1]],
        [bb[2], bb[1]],
        [bb[2], bb[3]],
        [bb[0], bb[3]],
        [bb[0], bb[1]],
    ]
    return coords


def coords_to_wkt_polygon(coords: list[float]) -> str:
    """Given a 2xN shaped list of coordinates, creates a WKT string.

    Parameters
    ----------
    coords : list[float]
        Coordinates to be converted.

    Returns
    -------
    str
        WKT string.
    """
    # Check if polygon is closed properly.
    if coords[0] != coords[-1]:
        coords.append(coords[0])
    merged_coords = ", ".join([f"{x[0]} {x[1]}" for x in coords])
    wkt = f"POLYGON (({merged_coords}))"
    return wkt


def wkt_polygon_to_coords(wkt: str) -> List[List[float]]:
    """Convert a WKT string for a `POLYGON` into a list of coordinates.

    Parameters
    ----------
    wkt : str
        `POLYGON` WKT string.

    Returns
    -------
    List[List[float]]
        List of coordinates.

    Raises
    ------
    ValueError
        Raised when a non-`POLYGON` WKT string is passed.
    """
    wkt_type = wkt.split(" ")[0]
    if wkt_type != "POLYGON":
        raise ValueError(
            f"Only wkt with geometry type `POLYGON` supported, not `{wkt_type}`."
        )
    coords_ = wkt.replace("POLYGON ((", "").replace("))", "").split(",")
    coords = [[float(y) for y in x.split(" ")] for x in coords_]
    # Check if polygon is closed properly.
    if coords[0] != coords[-1]:
        coords.append(coords[0])
    return coords


def to_vsimem(bb: List[float] = None, coords: List[List[float]] = None) -> str:
    """Convert a bounding-box to a ogr vector stored in `/vsimem/`.

    Parameters
    ----------
    bb : List[float]
        Bounding-box as [xmin, ymin, xmax, ymax].

    Returns
    -------
    str
        Path to geojson object in `/vsimem/`.
    """
    if bb is not None and coords is not None:
        raise ValueError  # TODO add test
    elif bb is not None:
        coords = bb_to_coords(bb)
    elif coords is not None:
        ...
    else:
        raise ValueError  # TODO add test

    wkt = coords_to_wkt_polygon(coords)

    geom = ogr.CreateGeometryFromWkt(wkt)
    outDriver = ogr.GetDriverByName("GeoJSON")
    fh = f"/vsimem/x_{time.strftime('%Y-%m-%d_%H%M%S')}_{np.random.randint(9999)}.geojson"
    outDataSource = outDriver.CreateDataSource(fh)
    outLayer = outDataSource.CreateLayer("temp", geom_type=ogr.wkbPolygon)
    featureDefn = outLayer.GetLayerDefn()
    outFeature = ogr.Feature(featureDefn)
    outFeature.SetGeometry(geom)
    outLayer.CreateFeature(outFeature)
    outFeature = None
    outDataSource = None

    return fh


#################
## ZONAL STATS ##
#################


def zone_stats(raster, wkt, wkt_srs):
    # Clip raster to the geometry.
    warp_options = gdal.WarpOptions(
        cutlineWKT=wkt,
        cropToCutline=True,
        cutlineSRS=wkt_srs,
        # creationOptions=["COMPRESS=LZW"],
        warpOptions=["CUTLINE_ALL_TOUCHED", "TRUE"],
    )
    output_file = (
        f"/vsimem/x_{time.strftime('%Y-%m-%d_%H%M%S')}_{np.random.randint(9999)}.tif"
    )
    ds: gdal.Dataset = gdal.Warp(output_file, raster, options=warp_options)

    # Get statistics for the clipped dataset.
    data = get_stats(ds)

    # Unlink.
    unlink_vsimems(output_file)

    return data


def zonal_stats(
    raster: str | gdal.Dataset,
    vector: str,
    identifier: str,
    n_threads: int = 1,
):
    unions, wkt_srs = group_geoms_on_attribute(vector, identifier, format="wkt")

    data_parts = []
    for name, wkt in unions.items():
        data = zone_stats(raster, wkt, wkt_srs)
        if isinstance(identifier, type(None)):
            identifier = "None"
        data[identifier] = name
        data_parts.append(data)

    # Calculate stats per ID value.
    # if n_threads <= 1:
    #     data_parts = map(
    #         zone_stats, repeat(raster), repeat(vector), repeat(identifier), id_values
    #     )
    # else:
    #     zone_stats_ = partial(zone_stats, raster, vector, identifier)
    #     executor = concurrent.futures.ThreadPoolExecutor(n_threads)
    #     futures = [executor.submit(zone_stats_, item) for item in id_values]
    #     concurrent.futures.wait(futures)
    #     data_parts = [x.result() for x in futures]

    data: pd.DataFrame = (
        pd.concat(data_parts)
        .dropna(subset=["mean", "minimum", "maximum"])
        .sort_values([identifier, "start_date"])
        .reset_index(drop=True)
    )

    # Add metadata to pd.DataFrame.
    data.attrs = make_df_md(raster)

    return data


def make_df_md(raster, include=["long_name", "units", "overview", "original_units"]):
    info = gdal.Info(raster, format="json")
    try:
        md: dict = info["bands"][0]["metadata"][""]
        assert isinstance(md, dict)
    except (KeyError, IndexError, AssertionError):
        logging.warning("No valid metadata found.")
        md = {}
    df_md = {k: v for k, v in md.items() if k in include}
    return df_md


if __name__ == "__main__":
    ...



import importlib.util
import logging
from string import ascii_lowercase, ascii_uppercase

import pandas as pd
from osgeo import gdal, gdalconst
from osgeo_utils import gdal_calc

optional_packages = ["dask", "rioxarray", "xarray"]
use_xarray = all(
    [not isinstance(importlib.util.find_spec(x), type(None)) for x in optional_packages]
)
if use_xarray:
    import xarray as xr


def __make_band_names__(length):
    letters = [x for x in ascii_lowercase + ascii_uppercase]
    i = 2
    while len(letters) < length:
        for letter in letters[:52]:
            letters.append(letter * i)
        i += 1
    return letters[:length]


def df_unit_convertor(df: pd.DataFrame, unit_conversion: str):

    valid_units = ["none", "dekad", "day", "month", "year"]
    if unit_conversion not in valid_units:
        raise ValueError(
            f"Please select one of {valid_units} instead of {unit_conversion}."
        )  # NOTE: TESTED
    
    source_unit = df.attrs["units"]

    source_unit_split = source_unit.split("/")
    source_unit_q = "/".join(source_unit_split[:-1])
    source_unit_time = source_unit_split[-1]

    conversion = {
        ("day", "day"): 1,
        ("day", "dekad"): df["number_of_days"].dt.days,
        ("day", "month"): df["start_date"].dt.days_in_month,
        ("day", "year"): 365,
        ("dekad", "day"): 1.0 / df["number_of_days"].dt.days,
        ("dekad", "month"): 3,
        ("dekad", "year"): 36,
        ("dekad", "dekad"): 1,
        ("month", "day"): 1 / df["start_date"].dt.days_in_month,
        ("month", "dekad"): 1 / 3,
        ("month", "month"): 1,
        ("month", "year"): 12,
        ("year", "dekad"): 1 / 36,
        ("year", "day"): 1 / 365,
        ("year", "month"): 1 / 12,
        ("year", "year"): 1,
    }.get((source_unit_time, unit_conversion), None)

    if isinstance(conversion, type(None)):
        logging.warning(
                "Couldn't succesfully determine unit conversion factors, keeping original units."
            )
        df.attrs["original_units"] = "N/A"
        df.attrs["units"] = source_unit
    else:
        df[["mean", "minimum", "maximum"]] = df[["mean", "minimum", "maximum"]].mul(
            conversion, axis=0
        )
        df.attrs["original_units"] = df.attrs["units"]
        df.attrs["units"] = f"{source_unit_q}/{unit_conversion}"

    return df


def unit_convertor(
    urls: list,
    in_fn: str,
    out_fn: str,
    unit_conversion: str,
    warp: gdal.Dataset,
    coptions=[],
) -> tuple:
    """Convert the units of multiple bands in a single geoTIFF file to another timescale.

    Parameters
    ----------
    urls : list
        Contains tuples of which the first item is a dictionary with metadata information for each band found in
        `in_fn`. Length of this list should be equal to the number of bands in `in_fn`.
    in_fn : str
        Path to geotiff file.
    out_fn : str
        Path to the to-be-created geotiff file.
    unit_conversion : str
        The desired temporal component of the converted units, should be one of
        "day", "dekad", "month" or "year".
    warp : gdal.Dataset
        The dataset to be adjusted, should point to `in_fn`.
    coptions : list, optional
        Extra creation options used to create `out_fn`, by default [].

    Returns
    -------
    tuple
        The new gdal.Dataset and the path to the created file.
    """

    global use_xarray

    input_files = dict()
    input_bands = dict()
    calc = list()
    should_convert = list()
    conversion_factors = list()
    letters = __make_band_names__(len(urls))

    if "AGERA5" in urls[0][1]:
        dtype = gdalconst.GDT_Float64
    else:
        dtype = gdalconst.GDT_Int32  # NOTE unit conversion can increase the DN's,
        # causing the data to not fit inside Int16 anymore...
        # so for now just moving up to Int32. Especially necessary
        # for NPP (which has a scale-factor of 0.001).

    for i, (md, _) in enumerate(urls):
        band_number = i + 1
        letter = letters[i]
        input_files[letter] = in_fn
        input_bands[f"{letter}_band"] = band_number
        if md.get("temporal_resolution", "unknown") == "Day":
            number_of_days = md.get("days_in_dekad", "unknown")
        else:
            number_of_days = md.get("number_of_days", "unknown")
        days_in_month = pd.Timestamp(md.get("start_date", "nat")).daysinmonth
        source_unit = md.get("units", "unknown")
        source_unit_split = source_unit.split("/")
        source_unit_q = "/".join(source_unit_split[:-1])
        source_unit_time = source_unit_split[-1]
        if any(
            [
                source_unit_time not in ["day", "month", "year", "dekad"],
                number_of_days == "unknown",
                source_unit == "unknown",
                pd.isnull(days_in_month),
            ]
        ):
            calc.append(f"{letter}.astype(numpy.float64)")
            md["units"] = source_unit
            md["units_conversion_factor"] = "N/A"
            md["original_units"] = "N/A"
            should_convert.append(False)
            conversion_factors.append(1)
        else:
            conversion = {
                ("day", "day"): 1,
                ("day", "dekad"): number_of_days,
                ("day", "month"): days_in_month,
                ("day", "year"): 365,
                ("dekad", "day"): 1 / number_of_days,
                ("dekad", "month"): 3,
                ("dekad", "year"): 36,
                ("dekad", "dekad"): 1,
                ("month", "day"): 1 / days_in_month,
                ("month", "dekad"): 1 / 3,
                ("month", "month"): 1,
                ("month", "year"): 12,
                ("year", "dekad"): 1 / 36,
                ("year", "day"): 1 / 365,
                ("year", "month"): 1 / 12,
                ("year", "year"): 1,
            }[(source_unit_time, unit_conversion)]
            calc.append(f"{letter}.astype(numpy.float64)*{conversion}")
            should_convert.append(True)
            conversion_factors.append(conversion)
            md["units"] = f"{source_unit_q}/{unit_conversion}"
            md["units_conversion_factor"] = conversion
            md["original_units"] = source_unit

    logging.debug(
        f"\ninput_files: {input_files}\ninput_bands: {input_bands}\ncalc: {calc}"
    )

    conversion_is_one = [x["units_conversion_factor"] == 1.0 for x, _ in urls]

    # NOTE See todo just below.
    scales = [warp.GetRasterBand(i + 1).GetScale() for i in range(warp.RasterCount)]
    offsets = [warp.GetRasterBand(i + 1).GetOffset() for i in range(warp.RasterCount)]

    logging.debug(f"\nSCALES: {scales}\nOFFSETS: {offsets}")

    if all(should_convert) and not all(conversion_is_one):
        logging.info(
            f"Converting units from [{source_unit}] to [{source_unit_q}/{unit_conversion}] (use_xarray = {use_xarray})."
        )

        ndv = warp.GetRasterBand(1).GetNoDataValue()
        if use_xarray:
            if "/vsi" not in in_fn:
                chunks = {"band": 1, "x": "auto", "y": "auto"}
            else:
                chunks = None
            ds = xr.open_dataset(
                in_fn,
                mask_and_scale=False,
                decode_coords="all",
                chunks=chunks,
            )
            xr_conv = xr.DataArray(conversion_factors, coords={"band": ds["band"]})
            ndv_ = ds["band_data"].attrs["_FillValue"]

            da = xr.where(
                ds["band_data"] == ndv_, ndv_, ds["band_data"] * xr_conv
            ).round(decimals=0)

            ds_out = da.to_dataset("band")
            for i, (scale, (md, _)) in enumerate(zip(scales, urls)):
                ds_out[i + 1].attrs = md
                ds_out[i + 1] = ds_out[i + 1].rio.write_nodata(ndv)
                ds_out[i + 1].attrs["scale_factor"] = scale

            ds_out = ds_out.rio.write_crs(ds.rio.crs)
            ds_out.rio.to_raster(
                out_fn,
                compress="LZW",
                dtype={5: "int32", 7: "float64"}[dtype],
                windowed=True,
                lock=True,
            )
            filen = out_fn
        else:
            logging.info(
                "Consider installing `xarray`, `rioxarray` and `dask` for faster unit conversions."
            )
            warp = gdal_calc.Calc(
                calc=calc,
                outfile=out_fn,
                overwrite=True,
                creation_options=coptions,
                quiet=True,
                type=dtype,
                NoDataValue=ndv,
                **input_files,
                **input_bands,
            )
            # TODO make bug report on GDAL for gdal_calc removing scale/offset factors
            for i, (scale, offset) in enumerate(zip(scales, offsets)):
                warp.GetRasterBand(i + 1).SetScale(scale)
                warp.GetRasterBand(i + 1).SetOffset(offset)
            warp.FlushCache()
            filen = out_fn
    else:
        if all(conversion_is_one):
            logging.info("Units are already as requested, no conversion needed.")
        else:
            logging.warning(
                "Couldn't succesfully determine unit conversion factors, keeping original units."
            )
        for i, (md, _) in enumerate(urls):
            if md["units_conversion_factor"] != "N/A":
                md["units"] = md["original_units"]
                md["units_conversion_factor"] = "N/A"
                md["original_units"] = "N/A"
        filen = in_fn

    return warp, filen



from typing import List, Tuple

from wapordl.toolbox.query import collect_responses


def generate_urls(
    variable: str,
    l3_region: str | None = None,
    period: List[str] | None = None,
    info: str = "downloadUrl",
) -> Tuple[str]:
    """Find resource URLs for an agERA5 variable for a specified period.

    Parameters
    ----------
    variable : str
        Name of the variable.
    l3_region : str | None, optional
        Three letter code specifying the level-3 region, by default None.
    period : list, optional
        Start and end date in between which resource URLs will be searched, by default None.
    info : str, optional
        Which items from the GISMGR to parse, options are ['workspaceCode',
        'mapsetCode',
        'code',
        'dimensions',
        'styleCode',
        'downloadUrl',
        'gsutilUri',
        'measureCaption',
        'measureUnit',
        'scale',
        'offset',
        'tilesSize',
        'overviewsResamplingAlgorithm',
        'bigTiff',
        'width',
        'height',
        'affineTransform',
        'srs',
        'extent',
        'dataType',
        'noDataValue',
        'flags',
        'classes',
        'size',
        'created',
        'updated',
        'updatesLog',
        'links']

    Returns
    -------
    tuple
        Resource URLs.

    Raises
    ------
    ValueError
        Invalid level selected.
    """

    level = variable.split("-")[0]

    if (level == "L1") or (level == "L2"):
        base_url = (
            "https://data.apps.fao.org/gismgr/api/v2/catalog/workspaces/WAPOR-3/mapsets"
        )
    elif level == "L3":
        base_url = "https://data.apps.fao.org/gismgr/api/v2/catalog/workspaces/WAPOR-3/mosaicsets"
    elif level == "AGERA5":
        base_url = (
            "https://data.apps.fao.org/gismgr/api/v2/catalog/workspaces/C3S/mapsets"
        )
    else:
        raise ValueError(f"Invalid level {level}.")  # NOTE: TESTED

    mapset_url = f"{base_url}/{variable}/rasters?filter="
    if not isinstance(l3_region, type(None)):
        mapset_url += f"code:CONTAINS:{l3_region};"
    if not isinstance(period, type(None)):
        mapset_url += f"time:OVERLAPS:{period[0]}:{period[1]};"

    urls = [x[0] for x in collect_responses(mapset_url, info=[info])]

    return tuple(sorted(urls))



import logging
import os

from osgeo import gdal

from wapordl.bounding_boxes import L3_BBS
from wapordl.toolbox import ogr_gdal
from wapordl.toolbox import query
from wapordl.products.wapor3 import generate_urls


def l3_bounding_boxes(variable="L3-T-A", l3_region=None) -> dict:
    """Determine the bounding-boxes of the WaPOR level-3 regions.

    Parameters
    ----------
    variable : str, optional
        Name of the variable used to check the bounding-box, by default "L3-T-A".
    l3_region : str, optional
        Name of the level-3 region to check, when `None` will check all available level-3 regions, by default None.

    Returns
    -------
    dict
        keys are three letter region codes, values are the coordinates of the bounding-boxes.
    """
    urls = generate_urls(
        variable, l3_region=l3_region, period=["2020-01-01", "2021-02-01"]
    )
    l3_bbs = {}
    for region_code, url in zip(
        [os.path.split(x)[-1].split(".")[-3] for x in urls], urls
    ):
        info = gdal.Info("/vsicurl/" + url, format="json")
        bb = info["wgs84Extent"]["coordinates"][0]
        l3_bbs[region_code] = bb
    return l3_bbs


def l3_codes() -> dict:
    url = 'https://data.apps.fao.org/gismgr/api/v2/catalog/workspaces/WAPOR-3/grids/L3-GRID/tiles'
    x = query.collect_responses(url, info=["code", "caption"])
    return {k:v for k,v in x}


def update_L3_BBS():
    """Add bounding-boxes to the global `L3_BBS` in case they are not yet hard-coded,
    in case new areas have come online since the last WaPORDL release.

    Returns
    -------
    list
        list with three letter codes for new areas.
    """
    logging.info("Updating L3 bounding-boxes.")
    all_l3_regions = l3_codes()
    new_regions = set(all_l3_regions.keys()).difference(set(L3_BBS.keys()))
    added_regions = list()
    for l3_region in new_regions:
        new_bb = l3_bounding_boxes(l3_region=l3_region).get(l3_region, None)
        if not isinstance(new_bb, type(None)):
            added_regions.append(l3_region)
            L3_BBS[l3_region] = new_bb
    return added_regions


def guess_l3_region(region_shape: str) -> str:
    """_summary_

    Parameters
    ----------
    region_shape : str
        Path to vector.

    Returns
    -------
    str
        Three letter code identifying the WaPOR level-3 region.

    Raises
    ------
    ValueError
        Raised when the `region_shape` doesn't intersect with any level-3 region.
    """
    checks = ogr_gdal.check_intersects(region_shape, L3_BBS)

    number_of_results = sum(checks.values())

    if number_of_results == 0:
        added_regions = update_L3_BBS()
        l3_bbs = {x: L3_BBS[x] for x in added_regions}
        checks = ogr_gdal.check_intersects(region_shape, l3_bbs)
        number_of_results = sum(checks.values())
        if number_of_results == 0:
            raise ValueError(
                "`region` can't be linked to any L3 region."
            )  # NOTE: TESTED

    l3_regions = [k for k, v in checks.items() if v]
    l3_region = l3_regions[0]
    if number_of_results > 1:
        logging.warning(
            f"`region` intersects with multiple L3 regions ({l3_regions}), continuing with {l3_region} only."
        )
    else:
        logging.info(f"Given `region` matches with `{l3_region}` L3 region.")

    return l3_region


if __name__ == "__main__":
    ...



from setuptools import find_packages, setup

setup(
    name="wapordl",
    version="1.2.1",
    packages=find_packages(include=["wapordl", "wapordl.*"]),
    include_package_data=True,
    python_requires=">=3.10",
    install_requires=[
        # "libgdal-netcdf", # conda-forge
        "requests",
        "pandas>=2.1.0,<3",
        "numpy>=1.15,<3",
        "gdal>=3.6.4,<4",
    ],
    extras_require={"full": ["xarray", "rioxarray", "dask", "matplotlib", "tqdm"]},
)



import logging
import os
from typing import List, Union

import numpy as np
import pandas as pd

import wapordl.products.agera5 as agera5
import wapordl.products.wapor3 as wapor3
import wapordl.toolbox.ogr_gdal as ogr_gdal
from wapordl.bounding_boxes import L2_BB, L3_BBS
from wapordl.overview_selector import determine_overview
from wapordl.region_selector import guess_l3_region, l3_bounding_boxes
from wapordl.toolbox.ts_queries import get_ts
from wapordl.unit_convertor import df_unit_convertor
from wapordl.variable_descriptions import collect_metadata, date_func


def parse_region(region: str | List[float] | None, level: str):
    """

    Parameters
    ----------
    region : Union[str, List[float], None]
        Defines the area of interest. Can be a three letter code to describe a WaPOR level-3 region,
        a path to a vector or raster file or a list of 4 floats, specifying a bounding box.
    level : str
        One of "L1", "L2", "L3" or "AGERA5".

    Returns
    -------
    tuple
        - The `l3_region` defines which L3 should be searched, its a three letter string or None.
        - The `region_code` is only used for naming purposes, e.g. in a filename, its a string.
        - The `region_shape` is a pathlike string that can be opened with GDAL.

    """
    global L3_BBS
    # L3-CODE
    if all([isinstance(region, str), len(region) == 3]):
        if not region == region.upper():
            raise ValueError(
                f"Invalid region code `{region}`, region codes have three capitalized letters."
            )

        if region not in list(L3_BBS.keys()):
            logging.info(f"Searching bounding-box for `{region}`.")
            bb = l3_bounding_boxes(l3_region=region)
            if len(bb) == 0:
                raise ValueError(f"Unkown L3 region `{region}`.")
            else:
                logging.info(f"Bounding-box found for `{region}`.")
                L3_BBS = {**L3_BBS, **bb}

        if level == "L3":
            l3_region = region[:]  # three letter code to filter L3 datasets in GISMGR2.
            region_code = l3_region[:]  # string to name the region in filenames etc.
            region_shape = None  # variable that can be passed to gdal.OpenEx(region_shape, gdal.OF_VECTOR)
        else:
            l3_region = None
            region_shape = ogr_gdal.to_vsimem(coords=L3_BBS[region])
            region_code = region[:]
    # GEOJSON
    elif isinstance(region, str) and ".tif" not in region:
        if "/vsicurl/" in region:
            region_code = "online_resource"
            region_shape = region
        elif not os.path.isfile(region):
            raise ValueError("Geojson file not found.")  # NOTE: TESTED
        else:
            region_code = os.path.split(region)[-1].replace(".geojson", "")
            epsg, driver, is_two_d = ogr_gdal.check_vector(region)
            if not np.all([epsg == 4326, driver == "GeoJSON", is_two_d]):
                ext_ = os.path.splitext(region)[-1]
                fn_ = os.path.split(region)[-1]
                out_fn_ = fn_.replace(ext_, "_reprojected.geojson")
                dim_ = {True: "2D", False: "3D"}[is_two_d]
                logging.warning(
                    f"Reprojecting `{fn_}` [EPSG:{epsg}, {dim_}] to `{out_fn_}` [EPSG:4326, 2D]."
                )
                region = ogr_gdal.reproject_vector(region, epsg=4326, in_memory=True)
            region_shape = region

        l3_region = None
    # BB or BB_from_geotiff
    elif isinstance(region, list) or (isinstance(region, str) and ".tif" in region):
        if isinstance(region, str) and ".tif" in region:
            region = ogr_gdal.get_wgs84_bounds(region)

        if not all([region[2] > region[0], region[3] > region[1]]):
            raise ValueError("Invalid bounding box.")  # NOTE: TESTED
        else:
            region_code = "bb"
            region_shape = ogr_gdal.to_vsimem(bb=region)
        l3_region = None
    else:
        raise ValueError(f"Invalid value for region ({region}).")  # NOTE: TESTED

    ## Check l3_region code.
    if level == "L3" and isinstance(l3_region, type(None)):
        l3_region = guess_l3_region(region_shape)
        region_code += f".{l3_region}"

    return l3_region, region_code, region_shape


def wapor_dl(
    region: str | List[float] | None,
    variable: str,
    period: List[str] = ["2021-01-01", "2022-01-01"],
    overview: str | int = "NONE",
    unit_conversion: str = "none",
    folder: str | None = None,
    filename: str | None = None,
    max_error: float = 0.5,
    make_plots: bool | str = False,
    warp_kwargs: dict = {},
) -> Union[str, pd.DataFrame]:
    """Download a WaPOR or agERA5 variable for a specified region and period.

    Parameters
    ----------
    region : Union[str, List[float], None]
        Defines the area of interest. Can be a three letter code to describe a WaPOR level-3 region,
        a path to a vector or raster file or a list of 4 floats, specifying a bounding box.
    variable : str
        Name of the variable to download.
    period : list, optional
        Period for which to download data, by default ["2021-01-01", "2022-01-01"].
    overview : str | int, optional
        Select which overview from the COGs to use. Can be "NONE" or -1 to use the original data,
        an integer from 0 up to the total of overviews available or "AUTO" for
        automatic overview detection, by default "NONE".
    unit_conversion : str, optional
        Apply a unit conversion on the created file, can be one of "none", "day", "dekad",
        "month" or "year", by default "none".
    folder : str, optional
        Path to a folder in which to save any (intermediate) files. If set to `None`, everything will be
        kept in memory, by default None.
    filename : str, optional
        Set a different name for the output file, by default None.
    max_error : float, optional
        Only used when `overview` is set to `"AUTO"`, sets the error threshold, by default 0.5.
    make_plots : bool | str, optional
        Only used when `overview` is set to `"AUTO"`, create graphs showing the iteration process
        executed to determine the optimal overview. When a existing folder is defined, the graphs
        are saved into that folder as png files, by default False.
    warp_kwargs : dict, optional
        Additional gdal.Warp keyword arguments used when downloading data, by default {}.

    Returns
    -------
    Union[str, pd.DataFrame]
        Return a path to a file (if `req_stats` is `None`) or a pd.Dataframe if req_stats is a list
        speciyfing statistics.
    """

    valid_units = ["none", "dekad", "day", "month", "year"]
    if unit_conversion not in valid_units:
        raise ValueError(
            f"Please select one of {valid_units} instead of {unit_conversion}."
        )  # NOTE: TESTED

    ## Retrieve info from variable name.
    level, _, tres = variable.split("-")

    l3_region, region_code, region_shape = parse_region(region, level)

    if isinstance(region_shape, type(None)):
        vsimems = list()
    elif "/vsimem/" in region_shape:
        vsimems = [region_shape]
    else:
        vsimems = list()

    ## Check the dates in period.
    if not isinstance(period, type(None)):
        period = [pd.Timestamp(x) for x in period]
        if period[0] > period[1]:
            raise ValueError("Invalid period.")  # NOTE: TESTED
        period = [x.strftime("%Y-%m-%d") for x in period]

    ## Collect urls for requested variable.
    if "AGERA5" in variable:
        urls = agera5.generate_urls(variable, period=period)
    else:
        urls = wapor3.generate_urls(variable, l3_region=l3_region, period=period)

    if len(urls) == 0:
        raise ValueError(
            "No files found for selected region, variable and period."
        )  # NOTE: TESTED

    ## Determine date for each url.
    md = collect_metadata(variable)
    md["overview"] = overview
    md_urls = [({**date_func(url, tres), **md}, url) for url in urls]

    logging.info(f"Found {len(md_urls)} files for {variable}.")

    # Determine overview
    if overview in ["auto", "AUTO"]:
        if isinstance(region_shape, type(None)):
            logging.warning(
                "Determining an overview level for an entire L3 region is unsupported, setting overview to `'NONE'`."
            )
            overview = "NONE"
        elif isinstance(region_shape, str) or isinstance(region_shape, list):
            logging.info("Searching for optimal overview.")
            overview = determine_overview(
                md_urls[0][1], region_shape, max_error=max_error, make_plots=make_plots
            )
            md_urls = [
                ({**md, **{"overview": f"AUTO:{overview}"}}, url) for md, url in md_urls
            ]
            logging.info(f"Using overview `{overview}`.")
        else:
            raise ValueError
    overview_ = -1 if overview == "NONE" else overview
    if overview not in [-1, "NONE"]:
        logging.warning("Downloading an overview instead of original data.")

    # Set gdal.Warp kwargs.
    info = ogr_gdal.get_info(md_urls[0][1])
    xres, yres = info["geoTransform"][1::4]
    warp_kwargs = {
        "xRes": abs(xres) * 2 ** (overview_ + 1),
        "yRes": abs(yres) * 2 ** (overview_ + 1),
        **warp_kwargs,
    }
    if isinstance(region_shape, str):
        warp_kwargs["cutlineDSName"] = region_shape

    ## Check if region overlaps with datasets bounding-box.
    if not isinstance(region_shape, type(None)) and level != "AGERA5":
        if level == "L2":
            bb_geom, bb_ftr = ogr_gdal.get_geom(L2_BB)[:2]
        else:
            data_bb_ = ogr_gdal.to_vsimem(coords=info["wgs84Extent"]["coordinates"][0])
            vsimems.append(data_bb_)
            bb_geom, bb_ftr = ogr_gdal.get_geom(data_bb_)[:2]
        region_geom, region_ftr = ogr_gdal.get_geom(region_shape)[:2]
        if not bb_geom.Intersects(region_geom):
            info_lbl1 = region_code if region_code != "bb" else str(region)
            info_lbl2 = (
                variable
                if isinstance(l3_region, type(None))
                else f"{variable}.{l3_region}"
            )
            raise ValueError(
                f"Selected region ({info_lbl1}) has no overlap with the datasets ({info_lbl2}) bounding-box."
            )
        del bb_ftr, region_ftr

    ## Determine output path.
    if folder:
        if not os.path.isdir(folder):
            os.makedirs(folder)
        if not isinstance(filename, type(None)):
            warp_fn = os.path.join(folder, f"{filename}.tif")
        else:
            overview__ = "NONE" if overview == -1 else overview
            warp_fn = os.path.join(
                folder, f"{region_code}_{variable}_{overview__}_{unit_conversion}.tif"
            )
    else:
        warp_fn = f"/vsimem/{pd.Timestamp.now()}_{region_code}_{variable}_{overview}_{unit_conversion}.tif"

    # Download the COG.
    warp_fn, vrt_fn = ogr_gdal.cog_dl(
        md_urls,
        warp_fn,
        overview=overview_,
        warp_kwargs=warp_kwargs,
        unit_conversion=unit_conversion,
    )
    # vsimems.append(warp_fn)
    vsimems.append(vrt_fn)

    ## Unlink memory files.
    ogr_gdal.unlink_vsimems(vsimems)

    return warp_fn


def wapor_map(
    region: Union[str, List[float], None],
    variable: str,
    period: List[str],
    folder: str,
    unit_conversion: str = "none",
    overview: str | int = "NONE",
    extension: str = ".tif",
    separate_unscale: bool = False,
    filename: str | None = None,
    max_error: float = 0.5,
    make_plots: bool | str = False,
    warp_kwargs: dict = {},
) -> str:
    """Download a map of a WaPOR3 or agERA5 variable for a specified region and period.

    Parameters
    ----------
    region : Union[str, List[float], None]
        Defines the area of interest. Can be a three letter code to describe a WaPOR level-3 region,
        a path to a vector file or a list of 4 floats, specifying a bounding box.
    variable : str
        Name of the variable to download.
    period : list
        Period for which to download data.
    folder : str
        Folder into which to download the data.
    unit_conversion : str, optional
        Apply a unit conversion on the created file, can be one of "none", "day", "dekad",
        "month" or "year", by default "none".
    overview : str | int, optional
        Select which overview from the COGs to use. Can be "NONE" or -1 to use the original data,
        an integer from 0 up to the total of overviews available or "AUTO" for
        automatic overview detection, by default "NONE".
    extension : str, optional
        One of ".tif" or ".nc", controls output format, by default ".tif".
    separate_unscale : bool, optional
        Set to `True` to create single band geotif files instead of a single geotif with multiple bands,
        does not do anything when extension is set to ".nc" , by default False.
    filename : str, optional
        Set a different name for the output file, by default None.
    max_error : float, optional
        Only used when `overview` is set to `"AUTO"`, sets the error threshold, by default 0.5.
    make_plots : bool | str, optional
        Only used when `overview` is set to `"AUTO"`, create graphs showing the iteration process
        executed to determine the optimal overview. When a existing folder is defined, the graphs
        are saved into that folder as png files, by default False.
    warp_kwargs : dict, optional
        Additional gdal.Warp keyword arguments used when downloading data, by default {}.

    Returns
    -------
    str
        Path to output file.
    """

    ## Check if a valid path to download into has been defined.
    if not os.path.isdir(folder):
        os.makedirs(folder)

    ## Call wapor_dl to create a GeoTIFF.
    fp = wapor_dl(
        region,
        variable,
        folder=folder,
        period=period,
        overview=overview,
        unit_conversion=unit_conversion,
        filename=filename,
        make_plots=make_plots,
        max_error=max_error,
        warp_kwargs=warp_kwargs,
    )

    fp = ogr_gdal.translate(fp, extension, separate_unscale=separate_unscale)

    ogr_gdal.unlink_vsimems(fp)

    return fp


def wapor_ts(
    region: Union[str, List[float], None],
    variable: str,
    period: List[str],
    method: str = "cloud",
    identifier: str | None = None,
    n_threads: int = 1,
    overview: str | int = "NONE",
    unit_conversion: str = "none",
    max_error: float = 0.5,
    make_plots: bool | str = False,
    warp_kwargs: dict = {},
) -> pd.DataFrame:
    """Download a timeseries of a WaPOR3 or agERA5 variable for a specified region and period.

    Parameters
    ----------
    region : Union[str, List[float], None]
        Defines the area of interest. Can be a three letter code to describe a WaPOR level-3 region,
        a path to a vector file or a list of 4 floats, specifying a bounding box.
    variable : str
        Name of the variable to download.
    period : list
        Period for which to download data.
    method : str
        Can be either "cloud" to calculate timeseries in the cloud or "local"
        to download the spatial data and calculate the timeseries locally.
    identifier : str | None, optional
        Choose an attribute name contained in the `region` vector file to use when
        calculating zonal statistics in order to create multiple timeseries, by default None.
    n_threads : int, optional
        Specify how many threads to use when calculating zonal statistics, by default 1. (Mutually
        exclusive with method='local').
    overview : str | int, optional
        Select which overview from the COGs to use. Can be "NONE" or -1 to use the original data,
        an integer from 0 up to the total of overviews available or "AUTO" for
        automatic overview detection, by default "NONE". (Mutually
        exclusive with method='local').
    unit_conversion : str, optional
        Apply a unit conversion on the created file, can be one of "none", "day", "dekad",
        "month" or "year", by default "none".
    max_error : float, optional
        Only used when `overview` is set to `"AUTO"`, sets the error threshold, by default 0.5. (Mutually
        exclusive with method='local').
    make_plots : bool | str, optional
        Only used when `overview` is set to `"AUTO"`, create graphs showing the iteration process
        executed to determine the optimal overview. When a existing folder is defined, the graphs
        are saved into that folder as png files, by default False. (Mutually
        exclusive with method='local').
    warp_kwargs : dict, optional
        Additional gdal.Warp keyword arguments used when downloading data, by default {}. (Mutually
        exclusive with method='local').

    Returns
    -------
    pd.DataFrame
        Timeseries output.
    """

    level = variable.split("-")[0]
    l3_region, _, region_shape = parse_region(region, level)

    if method == "local":
        fp = wapor_dl(
            region,
            variable,
            period=period,
            overview=overview,
            unit_conversion=unit_conversion,
            folder=None,
            max_error=max_error,
            make_plots=make_plots,
            warp_kwargs=warp_kwargs,
        )
        if (identifier is None) or (region_shape is None):
            df = ogr_gdal.get_stats(fp)
        else:
            df = ogr_gdal.zonal_stats(fp, region_shape, identifier, n_threads=n_threads)
        ogr_gdal.unlink_vsimems(fp)

    elif method == "cloud":
        df = get_ts(
            region_shape, period, variable, l3_region=l3_region, identifier=identifier
        )
        if isinstance(identifier, type(None)):
            df = df.drop("None", axis=1)
        if unit_conversion != "none":
            df = df_unit_convertor(df, unit_conversion)

    return df


if __name__ == "__main__":
    ...



import numpy as np
import pandas as pd
import requests

from wapordl.toolbox.ogr_gdal import group_geoms_on_attribute
from wapordl.variable_descriptions import AGERA5_VARS, WAPOR3_VARS


def get_ts(region_shape, period, variable, l3_region=None, identifier=None):
    
    if not isinstance(period, type(None)):
        period_ = [pd.Timestamp(x) for x in period]
        if period_[0] > period_[1]:
            raise ValueError("Invalid period.")  # NOTE: TESTED

    if "AGERA5" in variable:
        gsutiuri = AGERA5_VARS[variable]["gsutiuri"]
        scale_factor = AGERA5_VARS[variable]["scale"]
        long_name = AGERA5_VARS[variable]["long_name"]
        units = AGERA5_VARS[variable]["units"]
        overview = "NONE"
    else:
        gsutiuri = WAPOR3_VARS[variable]["gsutiuri"]
        scale_factor = WAPOR3_VARS[variable]["scale"]
        long_name = WAPOR3_VARS[variable]["long_name"]
        units = WAPOR3_VARS[variable]["units"]
        overview = "NONE"

    unions, _ = group_geoms_on_attribute(region_shape, identifier, format="json")

    if "L3" in variable:
        if isinstance(l3_region, type(None)):
            raise ValueError("Please specify a valid three letter `l3_region`.")
        gsutiuri += f".{l3_region}"

    data_parts = []
    for name, json_string in unions.items():
        data_ = request_ts(json_string, period, gsutiuri)
        data = parse_data(data_, scale_factor)
        if isinstance(identifier, type(None)):
            identifier = "None"
        data[identifier] = name
        data_parts.append(data)

    data: pd.DataFrame = (
        pd.concat(data_parts)
        .dropna(subset=["mean", "minimum", "maximum"])
        .sort_values([identifier, "start_date"])
        .reset_index(drop=True)
    )

    data["number_of_days"] = (data["end_date"] - data["start_date"]) + pd.Timedelta(
        days=1
    )

    data.attrs = {"long_name": long_name, "units": units, "overview": overview}

    return data


def request_ts(json_string, period, gsutiuri, extra_post_args={"all_touched": True}):
    to_post = {
        "GEE_imageCollection_id": gsutiuri,
        "area_of_interest": eval(json_string),
        "dateRange": {"startDate": period[0], "endDate": period[1]},
        **extra_post_args,
    }
    endpoint = "https://api.data.apps.fao.org/api/v2/map/zonalstats?streaming=false"
    resp = requests.post(endpoint, json=to_post)
    resp.raise_for_status()
    return resp.json()


def parse_data(data, scale_factor):
    df_data = {
        "start_date": [],
        "end_date": [],
        "minimum": [],
        "maximum": [],
        "mean": [],
    }
    for item in data:
        df_data["start_date"].append(np.datetime64(item["system_time_start"], "ms"))
        df_data["end_date"].append(np.datetime64(item["system_time_end"], "ms"))

        if item["min"] is not None:
            df_data["minimum"].append(item["min"] * scale_factor)
        else:
            df_data["minimum"].append(pd.NA)

        if item["max"] is not None:
            df_data["maximum"].append(item["max"] * scale_factor)
        else:
            df_data["maximum"].append(pd.NA)

        if item["mean"] is not None:
            df_data["mean"].append(item["mean"] * scale_factor)
        else:
            df_data["mean"].append(pd.NA)

    df = pd.DataFrame(df_data)
    return df


if __name__ == "__main__":
    ...
